#    Copyright (C) 2016 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""N dimensional datasets."""

import numpy as N

from .. import utils

from .commonfn import _, dsPreviewHelper
from .base import DatasetConcreteBase

class DatasetNDBase(DatasetConcreteBase):
    """N-dimensional datasets."""

    dimensions = -1
    dstype = _('nD')
    editable = False

    def userSize(self):
        return '×'.join(str(x) for x in self.data.shape)

    def userPreview(self):
        return dsPreviewHelper(N.ravel(self.data))

    def description(self):
        return _('ND (%s), numeric') % self.userSize()

    def returnCopy(self):
        return DatasetND(data=self.data)

    def returnCopyWithNewData(self, **args):
        return DatasetND(**args)

    def empty(self):
        """Is the data defined?"""
        return len(self.data) == 0

    def datasetAsText(self, fmt='%g', join='\t'):
        """Dataset as text for copy, paste, etc."""
        def fmtrecurse(arr):
            if arr.ndim == 0:
                return fmt % arr + '\n'
            elif arr.ndim == 1:
                out = [fmt % v for v in arr]
                return join.join(out)
            else:
                out = []
                for v in arr:
                    out.append(fmtrecurse(v))
                out.append('')
                return '\n'.join(out)

        return fmtrecurse(self.data)

class DatasetND(DatasetNDBase):
    def __init__(self, data=None):
        """data is a numpy array of N dimensions."""

        DatasetNDBase.__init__(self)

        if isinstance(data, N.ndarray):
            self.data = data.astype(N.float64)
        elif isinstance(data, list) or isinstance(data, tuple):
            self.data = N.array(dtype=N.float64)
        else:
            raise ValueError("Could not convert data to nD numpy array.")

    def saveDataDumpToText(self, fileobj, name):
        """Save data to vsz in form of text."""

        fileobj.write("ImportStringND(%s, '''\n" % repr(name))
        if self.data.shape[0] == 1:
            # unfortunately it's hard to decode a single dimension
            # here so we record this unambiguously
            shape = ' '.join((str(d) for d in self.data.shape))
            fileobj.write("shape %s\n" % shape)
        fileobj.write(self.datasetAsText(fmt='%e', join=' '))
        fileobj.write("''')\n")

    def saveDataDumpToHDF5(self, group, name):
        """Save dataset to VSZ HDF5 format."""

        escname = utils.escapeHDFDataName(name)
        group[escname] = self.data
        group[escname].attrs['vsz_datatype'] = 'nd'
        group[escname].attrs['vsz_name'] = name.encode('utf-8')



#    Copyright (C) 2020 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""For histogramming data."""

import numpy as N

from .. import qtall as qt
from .. import document
from .. import setting
from .. import utils
from ..helpers import qtloops

from .plotters import GenericPlotter
from .point import ErrorBarDraw

def _(text, disambiguation=None, context='Histo'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

scaling_fwd = {
    'linear': lambda x: N.array(x),
    'sqrt': lambda x: N.sqrt(x),
    'log': lambda x: N.log(x),
    'arcsinh': lambda x: N.arcsinh(x),
    'exp': lambda x: N.exp(x),
    'sqr': lambda x: x**2,
    'sinh': lambda x: N.sinh(x),
}

scaling_bkd = {
    'linear': lambda x: N.array(x),
    'sqrt': lambda x: x**2,
    'log': lambda x: N.exp(x),
    'arcsinh': lambda x: N.sinh(x),
    'exp': lambda x: N.log(x),
    'sqr': lambda x: N.sqrt(x),
    'sinh': lambda x: N.arcsinh(x),
}

def _calcerrs(data, errmode):
    if errmode == 'sqrt':
        err = N.sqrt(data)
        return err, -err
    elif errmode == 'gehrels':
        perr = 1 + N.sqrt(data + 0.75)
        nerr = -N.sqrt(N.clip(data - 0.25, 0, None))
        return perr, nerr
    elif errmode == 'none':
        return None, None
    else:
        raise RuntimeError('Unknown error mode')

def _specialbin(mode, data, edges):
    """Special binning modes."""
    func = {
        'stddev': N.std,
        'variance': N.var,
        'mean': N.mean,
        'median': N.median,
        'min': N.amin,
        'max': N.amax,
    }[mode]

    data = data[N.isfinite(data)]
    idxs = N.searchsorted(edges, data)-1
    out = N.empty(len(edges)-1)
    for i in range(len(edges)-1):
        dbin = data[idxs==i]
        if len(dbin)==0:
            out[i] = N.nan
        else:
            out[i] = func(dbin)
    return out

def doBinning(data, weights=None,
              scaling='linear', minval='Auto', maxval='Auto',
              mode='constant', numbins=10, manualbins=None,
              calcmode='counts', errormode='gehrels',
):

    if weights:
        minlen = min(len(weights), len(data))
        weights = weights.data[:minlen]
        data = data[:minlen]

    # scale data according to function
    sfwd = scaling_fwd[scaling]
    sbkd = scaling_bkd[scaling]
    sdata = sfwd(data)

    if minval=='Auto' or minval is None:
        minval = N.nanmin(sdata)
    else:
        minval = sfwd(minval)

    if maxval=='Auto' or maxval is None:
        maxval = N.nanmax(sdata)
    else:
        maxval = sfwd(maxval)

    if not N.isfinite(minval) or not N.isfinite(maxval) or maxval<=minval:
        # problem with scaling or values
        return

    # non finite bins should be included in fractions, etc.
    sdata[~N.isfinite(sdata)] = N.inf

    if mode == 'constant':
        bins = numbins
    elif mode == 'manual':
        bins = N.unique(manualbins)
        if len(bins)<2:
            return
        bins = sfwd(bins)
        bins = bins[N.isfinite(bins)]
    else:
        bins = mode

    hist, edges = N.histogram(
        sdata,
        weights=weights,
        range=(minval, maxval),
        bins=bins
    )

    # scale edges back after transformation
    sedges = sbkd(edges)

    if calcmode in {'counts-cumulative', 'fraction-cumulative'}:
        hist = N.cumsum(hist)
    elif calcmode in {'counts-cumulative-reverse', 'fraction-cumulative-reverse'}:
        hist = N.cumsum(hist[::-1])[::-1]

    # special calculation modes
    perr = nerr = None
    if calcmode in {'stddev', 'variance', 'mean', 'median', 'min', 'max'}:
        hist = _specialbin(calcmode, sdata, edges)
    elif calcmode in {
            'counts', 'fraction', 'density', 'density_scaled',
            'counts-cumulative', 'counts-cumulative-reverse',
            'fraction-cumulative', 'fraction-cumulative-reverse',
    } and weights is None:
        perr, nerr = _calcerrs(hist, errormode)

    if calcmode == 'density' or calcmode == 'density_scaled':
        e = sedges if calcmode=='density' else edges
        histscale = (1/hist.sum()) / (e[1:]-e[:-1])
        hist = hist * histscale
        if perr is not None:
            perr = perr * histscale
        if nerr is not None:
            nerr = nerr * histscale

    if calcmode in {
            'fraction', 'fraction-cumulative', 'fraction-cumulative-reverse'}:
        invcts = 1/len(data)
        hist = hist * invcts
        if perr is not None:
            perr *= invcts
        if nerr is not None:
            nerr *= invcts

    return hist, perr, nerr, sedges

class FillBrush1(setting.BrushExtended):
    def __init__(self, name, **args):
        setting.BrushExtended.__init__(self, name, **args)
        self.get('color').newDefault( setting.Reference('../color') )
        self.add( setting.Choice(
            'mode',
            ('under', 'over', 'tozero'),
            'tozero',
            descr=_('Mode'),
            usertext=_('Mode')), 0)
        self.add( setting.Bool(
            'hideerror', False,
            descr=_('Hide the filled region inside the error bars'),
            usertext=_('Hide error fill')) )

class FillBrush2(FillBrush1):
    def __init__(self, name, **args):
        FillBrush1.__init__(self, name, **args)
        self.get('hide').newDefault(True)
        self.get('mode').newDefault('over')

class PlotLine(setting.Line):
    def __init__(self, name, **args):
        setting.Line.__init__(self, name, **args)
        self.get('color').newDefault( setting.Reference('../color') )

class PostLine(setting.Line):
    def __init__(self, name, **args):
        setting.Line.__init__(self, name, **args)
        self.get('hide').newDefault(True)
        self.get('color').newDefault( setting.Reference('../color') )

class MarkerFillBrush(setting.Brush):
    def __init__(self, name, **args):
        setting.Brush.__init__(self, name, **args)
        self.get('color').newDefault( setting.Reference('../color') )

class Histo(GenericPlotter):
    """Widget for histogramming data."""

    typename='histo'
    allowusercreation=True
    description=_('Histogram of a dataset')

    def __init__(self, *args, **argsv):
        GenericPlotter.__init__(self, *args, **argsv)
        self.changeset = -1
        self.hist = self.edges = self.perrs = self.nerrs = None

    @staticmethod
    def _showbinning(mode):
        # which bins to show depending on binning mode
        if mode == 'constant':
            return ('numbins',), ('manual',)
        elif mode == 'manual':
            return ('manual',), ('numbins',),
        else:
            # named mode
            return (), ('manual', 'numbins')

    @classmethod
    def addSettings(klass, s):
        GenericPlotter.addSettings(s)

        s.add( setting.DatasetExtended(
            'data', '',
            descr=_('Dataset to apply binning to'),
            usertext=_('Bin dataset')), 1 )

        s.add( setting.DatasetExtended(
            'weights', '',
            descr=_('Optional weight applied to counts of data'),
            usertext=_('Weights')), 3 )

        s.add( setting.Choice(
            'calcmode',
            (
                'counts', 'fraction', 'density',
                'counts-cumulative', 'counts-cumulative-reverse',
                'fraction-cumulative', 'fraction-cumulative-reverse',
            ),
            'counts',
            descr=_('Calculate when binning'),
            usertext=_('Calculate')), 4 )

        s.add( setting.ChoiceSwitch(
            'binning',
            (
                'constant', 'manual', 'auto',
                'fd', 'doane', 'scott', 'stone', 'rice',
                'sturges', 'sqrt'
            ),
            'constant',
            showfn=klass._showbinning,
            descr=_('Binning mode'),
            usertext=_('Binning')), 5 )

        s.add( setting.FloatOrAuto(
            'minval', 'Auto',
            descr=_('Minimum of range'),
            usertext=_('Minimum')), 6 )
        s.add( setting.FloatOrAuto(
            'maxval', 'Auto',
            descr=_('Maximum of range'),
            usertext=_('Maximum')), 7 )

        s.add( setting.ChoiceSwitch(
            'scaling',
            ('linear', 'log', 'sqrt', 'arcsinh', 'exp', 'sqr', 'sinh'),
            'linear',
            descr=_('Data scaling before creating bins'),
            usertext=_('Bin scaling')), 9 )
        s.add( setting.Int(
            'numbins', 10,
            minval=1, maxval=100000,
            descr=_('Number of bins'),
            usertext=_('Number')), 9 )
        s.add( setting.FloatList(
            'manual',
            [],
            descr=_('Manual binning edges'),
            usertext=_('Manual')), 10 )

        s.add( setting.Choice(
            'errormode',
            ('none', 'sqrt', 'gehrels'),
            'gehrels',
            descr=_('Error estimation'),
            usertext=_('Uncertainty')), 11 )

        s.add( setting.Choice(
            'direction',
            ('horizontal', 'vertical'),
            'vertical',
            descr=_('Bars direction'),
            usertext=_('Direction')), 12 )

        s.add( setting.Color(
            'color',
            'auto',
            descr=_('Master color'),
            usertext=_('Color'),
            formatting=True), 0 )
        s.add( setting.Choice(
            'style',
            ('step', 'join'),
            'step',
            descr=_('Drawing style'),
            usertext=_('Style'),
            formatting=True), 1 )
        s.add( setting.Marker(
            'marker',
            'none',
            descr=_('Type of marker to plot'),
            usertext=_('Marker'), formatting=True), 2 )
        s.add( setting.DistancePt(
            'markerSize',
            '3pt',
            descr=_('Size of marker to plot'),
            usertext=_('Marker size'), formatting=True), 3 )
        s.add( setting.ErrorStyle(
            'errorStyle',
            'none',
            descr=_('Style of error bars to plot'),
            usertext=_('Error style'), formatting=True), 4 )

        s.add( PlotLine(
            'Line',
            descr=_('Plot line'),
            usertext=_('Plot line')),
            pixmap='settings_plotline' )
        s.add( FillBrush1(
            'Fill1',
            descr=_('Fill under'),
            usertext=_('Fill under')),
            pixmap='settings_plotfillbelow' )
        s.add( FillBrush2(
            'Fill2',
            descr=_('Fill over'),
            usertext=_('Fill over')),
            pixmap='settings_plotfillabove' )
        s.add( PostLine(
            'PostLine',
            descr=_('Post line'),
            usertext=_('Post line')),
            pixmap='settings_postline' )

        s.add( setting.MarkerLine(
            'MarkerLine',
            descr=_('Line around marker'),
            usertext=_('Marker border')),
            pixmap='settings_plotmarkerline' )
        s.add( MarkerFillBrush(
            'MarkerFill',
            descr=_('Marker fill'),
            usertext=_('Marker fill')),
            pixmap='settings_plotmarkerfill' )

        s.add( setting.ErrorBarLine(
            'ErrorBarLine',
            descr=_('Error bar line'),
            usertext=_('Error bar line')),
            pixmap='settings_ploterrorline' )
        s.ErrorBarLine.get('color').newDefault( setting.Reference('../color') )

    def computeHisto(self):
        if self.document.changeset < self.changeset:
            return
        self.changeset = self.document.changeset

        s = self.settings
        self.hist = self.edges = self.perrs = self.nerrs = None

        dsetn = self.settings.get('data')
        ds = dsetn.getData(self.document)
        if ds is None or dsetn.isEmpty():
            return
        data = ds.data
        if len(data) == 0:
            return

        # weighting for points
        weights = self.settings.get('weights').getData(self.document)
        if not weights:
            weights = None

        retn = doBinning(
            data, weights=weights,
            scaling=s.scaling, minval=s.minval, maxval=s.maxval,
            mode=s.binning, numbins=s.numbins,
            manualbins=s.manual,
            calcmode=s.calcmode, errormode=s.errormode)
        if retn is None:
            return

        self.hist, self.perr, self.nerr, self.edges = retn

    def affectsAxisRange(self):
        """Which axes are affected by this plotter?"""
        s = self.settings
        if s.direction == 'vertical':
            return ( (s.xAxis, 'bins'), (s.yAxis, 'cts') )
        else:
            return ( (s.xAxis, 'cts'), (s.yAxis, 'bins') )

    def getRange(self, axis, depname, axrange):
        """Update axis range based on data."""
        self.computeHisto()
        if self.hist is None:
            return

        if depname == 'bins':
            # we're looking at the bin positions
            dlo = dhi = self.edges

        else:
            # these are the bin heights
            dhi = dlo = self.hist
            if self.perr is not None:
                dhi = dhi + self.perr
            if self.nerr is not None:
                dlo = dlo + self.nerr

        minv = maxv = None
        if axis.settings.log:
            sel = dlo>0
            if N.any(sel):
                minv = dlo[sel].min()
            sel = dhi>0
            if N.any(sel):
                maxv = dhi[sel].max()
        else:
            if len(dlo) > 0:
                minv = dlo.min()
            if len(dhi) > 0:
                maxv = dhi.max()

        if minv is not None:
            axrange[0] = min(axrange[0], minv)
        if maxv is not None:
            axrange[1] = max(axrange[1], maxv)

    def drawKeySymbol(self, number, painter, x, y, width, height):
        """Draw the line fill for the key."""

        s = self.settings

        # draw fill setting
        yp = y + height/2
        for fill in s.Fill1, s.Fill2:
            if not fill.hide:
                if fill.mode =='over':
                    y2 = yp-height*0.45
                else:
                    y2 = yp+height*0.45
                path = qt.QPainterPath()
                path.addRect(qt.QRectF(
                    qt.QPointF(x, yp), qt.QPointF(x+width, y2)))
                utils.brushExtFillPath(painter, fill, path)

        # draw line
        if not s.Line.hide:
            painter.setPen( s.Line.makeQPen(painter) )
            painter.drawLine( qt.QPointF(x, yp), qt.QPointF(x+width, yp) )

    def dataDraw(self, painter, axes, posn, cliprect):
        """Draw the histogram."""

        s = self.settings
        self.computeHisto()
        if self.hist is None:
            return

        vert = s.direction == 'vertical'
        if vert:
            binaxis, ctaxis = axes
        else:
            ctaxis, binaxis = axes
        ctslog = ctaxis.settings.log

        # calculate bin midpoint according to scaling
        scaling = s.scaling
        sedges = scaling_fwd[scaling](self.edges)
        midpts = scaling_bkd[scaling]( 0.5*(sedges[1:]+sedges[:-1]) )
        midplot = binaxis.dataToPlotterCoords(posn, midpts)

        linepoly = qt.QPolygonF()
        ctsplot = ctaxis.dataToPlotterCoords(posn, self.hist)
        style = s.style
        if style == 'step':
            # stepped line
            edgeplot = binaxis.dataToPlotterCoords(posn, self.edges)
            if vert:
                qtloops.addNumpyToPolygonF(
                    linepoly, edgeplot[:-1], ctsplot, edgeplot[1:], ctsplot)
            else:
                qtloops.addNumpyToPolygonF(
                    linepoly, ctsplot, edgeplot[:-1], ctsplot, edgeplot[1:])
        else:
            # non-stepped line
            if vert:
                qtloops.addNumpyToPolygonF(linepoly, midplot, ctsplot)
            else:
                qtloops.addNumpyToPolygonF(linepoly, ctsplot, midplot)

        # coordinates of zero value
        if ctslog:
            zeropos = ctaxis.dataToPlotterCoords(posn, N.array([1e-99]))[0]
        else:
            zeropos = ctaxis.dataToPlotterCoords(posn, N.array([0]))[0]

        # fill above or below the curve
        for fill in s.Fill1, s.Fill2:
            if not fill.hide:
                firstpt = linepoly.first()
                lastpt = linepoly.last()
                fillpoly = qt.QPolygonF(linepoly)
                if vert:
                    if fill.mode == 'under':
                        yfill = posn[3]
                    elif fill.mode == 'over':
                        yfill = posn[1]
                    elif fill.mode == 'tozero':
                        yfill = zeropos
                    fillpoly.append(qt.QPointF(lastpt.x(), yfill))
                    fillpoly.append(qt.QPointF(firstpt.x(), yfill))
                else:
                    if fill.mode == 'under':
                        xfill = posn[0]
                    elif fill.mode == 'over':
                        xfill = posn[2]
                    elif fill.mode == 'tozero':
                        xfill = zeropos
                    fillpoly.append(qt.QPointF(xfill, lastpt.y()))
                    fillpoly.append(qt.QPointF(xfill, firstpt.y()))
                clippoly = qt.QPolygonF()
                qtloops.polygonClip(fillpoly, cliprect, clippoly)
                path = qt.QPainterPath()
                path.addPolygon(clippoly)
                utils.brushExtFillPath(painter, fill, path)

        # the post line which goes down to zero from the data line
        if not s.PostLine.hide:
            painter.setPen(s.PostLine.makeQPen(painter))
            pos = edgeplot if style=='step' else midplot
            zeros = N.full(pos.shape, zeropos)
            if vert:
                qtloops.plotLinesToPainter(
                    painter, pos, ctsplot, pos, zeros,
                    cliprect)
            else:
                qtloops.plotLinesToPainter(
                    painter, ctsplot, pos, zeros, pos,
                    cliprect)

        # the plot line (stepped or not)
        if not s.Line.hide:
            painter.setPen(s.Line.makeQPen(painter))
            qtloops.plotClippedPolyline(painter, cliprect, linepoly)

        # points and error bars
        if vert:
            xplt, yplt = midplot, ctsplot
        else:
            xplt, yplt = ctsplot, midplot

        # any error bars
        markersize = s.get('markerSize').convert(painter)
        if s.errorStyle != 'none' and self.perr is not None:
            minv = ctaxis.dataToPlotterCoords(posn, self.hist+self.nerr)
            maxv = ctaxis.dataToPlotterCoords(posn, self.hist+self.perr)
            if vert:
                ymin, ymax = minv, maxv
                xmin = xmax = None
            else:
                xmin, xmax = minv, maxv
                ymin = ymax = None

            ebp = ErrorBarDraw(
                s.errorStyle, s.ErrorBarLine, s.Fill2, s.Fill1, markersize)
            ebp.plot(painter, xmin, xmax, ymin, ymax, xplt, yplt, cliprect)

        # plot data points
        if not s.MarkerLine.hide or not s.MarkerFill.hide:
            if not s.MarkerFill.hide:
                painter.setBrush(s.MarkerFill.makeQBrush(painter))
            else:
                painter.setBrush(qt.QBrush())

            if not s.MarkerLine.hide:
                painter.setPen(s.MarkerLine.makeQPen(painter))
            else:
                painter.setPen(qt.QPen(qt.Qt.NoPen))

            utils.plotMarkers(
                painter, xplt, yplt, s.marker, markersize,
                clip=cliprect)

document.thefactory.register(Histo)



# commandinterpreter.py
# this module handles the command line interface interpreter

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""
A module for the execution of user 'macro' code inside a special
environment. That way the commandline can be used to interact with
the app without worrying about the app internals.

The way this works is to create an evironment specific to the class
consisting of globals & locals.

Commands can be run inside the environment. Python errors are trapped
and dumped out to stderr.

stderr and stdout can be reassigned in the environment to point to
an alternative interface. They should point to objects providing
a write() interface.

This class is modelled on the one described in
'GUI Programming in Python: QT Edition' (Boudewijn Rempt)
"""

# get globals before things are imported
_globals = globals()

import sys
import traceback
import io
import os.path
import pickle

from .commandinterface import CommandInterface
from .. import utils

class CommandInterpreter:
    """Class for executing commands in the Veusz command line language."""

    def __init__(self, document):
        """ Initialise object with the document it interfaces."""
        self.document = document

        # set up interface to document
        self.interface = CommandInterface(document)

        # initialise environment (make a copy from inital globals)
        self.globals = _globals.copy()

        # save the stdout & stderr
        self.write_stdout = sys.stdout
        self.write_stderr = sys.stderr
        self.read_stdin = sys.stdin

        # import numpy into the environment
        exec("from numpy import *", self.globals)

        # define root object
        self.globals['Root'] = self.interface.Root

        # shortcut
        ifc = self.interface

        # define commands for interface
        self.cmds = {}
        for cmd in (
                CommandInterface.safe_commands +
                CommandInterface.unsafe_commands +
                CommandInterface.import_commands):
            self.cmds[cmd] = getattr(ifc, cmd)
        self.cmds['GPL'] = self.GPL
        self.cmds['Load'] = self.Load

        self.globals.update( self.cmds )

    def addCommand(self, name, command):
        """Add the given command to the list of available commands."""
        self.cmds[name] = command
        self.globals[name] = command

    def setFiles(self, stdout, stderr, stdin):
        """Assign the environment input/output files."""
        self.write_stdout = stdout
        self.write_stderr = stderr
        self.read_stdin = stdin

    def _pythonise(self, text):
        """Internal routine to convert commands in the form Cmd a b c into Cmd(a,b,c)."""

        out = ''
        # iterate over lines
        for line in text.split('\n'):
            parts = line.split()

            # turn Cmd a b c into Cmd(a,b,c)
            if len(parts) != 0 and parts[0] in self.cmds:
                line = utils.pythonise(line)

            out += line + '\n'

        return out

    def run(self, inputcmds, filename = None):
        """ Run a set of commands inside the preserved environment.

        inputcmds: a string with the commands to run
        filename: a filename to report if there are errors
        """

        if filename is None:
            filename = '<string>'

        # pythonise!
        inputcmds = self._pythonise(inputcmds)

        # ignore if blank
        if len(inputcmds.strip()) == 0:
            return

        # preserve output streams
        saved = sys.stdout, sys.stderr, sys.stdin
        sys.stdout, sys.stderr, sys.stdin = (
            self.write_stdout, self.write_stderr, self.read_stdin)

        # count number of newlines in expression
        # If it's 2, then execute as a single statement (print out result)
        if inputcmds.count('\n') == 2:
            stattype = 'single'
        else:
            stattype = 'exec'

        # first compile the code to check for syntax errors
        try:
            c = compile(inputcmds, filename, stattype)
        except (OverflowError, ValueError, SyntaxError):
            info = sys.exc_info()
            backtrace = traceback.format_exception(*info)
            for line in backtrace:
                sys.stderr.write(line)

        else:
            # block update signals from document while updating
            with self.document.suspend():
                try:
                    # execute the code
                    exec(c, self.globals)
                except:
                    # print out the backtrace to stderr
                    info = sys.exc_info()
                    backtrace = traceback.format_exception(*info)
                    for line in backtrace:
                        sys.stderr.write(line)

        # return output streams
        sys.stdout, sys.stderr, sys.stdin = saved

    def Load(self, filename):
        """Replace the document with a new one from the filename."""

        with io.open(filename, 'r', encoding='utf8') as f:
            self.document.wipe()
            self.interface.To('/')
            oldfile = self.globals['__file__']
            absfname = os.path.abspath(filename)
            self.globals['__file__'] = absfname

            self.interface.importpath.append(
                os.path.dirname(os.path.abspath(filename)))
            self.runFile(f)
            self.interface.importpath.pop()
            self.globals['__file__'] = oldfile
            self.document.filename = absfname
            self.document.setModified()
            self.document.setModified(False)
            self.document.clearHistory()

    def runFile(self, fileobject):
        """ Run a file in the preserved environment."""

        # preserve output streams
        temp_stdout = sys.stdout
        temp_stderr = sys.stderr
        sys.stdout = self.write_stdout
        sys.stderr = self.write_stderr

        with self.document.suspend():
            # actually run the code
            try:
                exec(fileobject.read(), self.globals)
            except Exception:
                # print out the backtrace to stderr
                info = sys.exc_info()
                backtrace = traceback.format_exception(*info)
                for line in backtrace:
                    sys.stderr.write(line)

        # return output streams
        sys.stdout = temp_stdout
        sys.stderr = temp_stderr

    def evaluate(self, expression):
        """Evaluate an expression in the environment."""

        # preserve output streams
        temp_stdout = sys.stdout
        temp_stderr = sys.stderr
        sys.stdout = self.write_stdout
        sys.stderr = self.write_stderr

        # actually run the code
        try:
            retn = eval(expression, self.globals)
        except Exception:
            # print out the backtrace to stderr
            info = sys.exc_info()
            backtrace = traceback.format_exception(*info)
            for line in backtrace:
                sys.stderr.write(line)
            retn = None

        # return output streams
        sys.stdout = temp_stdout
        sys.stderr = temp_stderr

        return retn

    def GPL(self):
        """Write the GPL to the console window."""
        sys.stdout.write( utils.getLicense() )

    def runPickle(self, command):
        """Run a pickled command given as arguments.

        command should consist of following:
        dumps( (name, args, namedargs) )
        name is name of function to execute in environment
        args are the arguments (list)
        namedargs are the named arguments (dict).
        """

        name, args, namedargs = pickle.loads(command)
        self.globals['_tmp_args0'] = args
        self.globals['_tmp_args1'] = namedargs

        #print(name, args, namedargs)
        try:
            retn = eval('%s(*_tmp_args0, **_tmp_args1)' % name)
        except Exception as e:
            # return exception picked if exception
            retn = e

        del self.globals['_tmp_args0']
        del self.globals['_tmp_args1']

        return pickle.dumps(retn)



#    Copyright (C) 2014 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""Generic 3D plotting widet."""

from .. import qtall as qt

from .. import setting
from . import widget
from ..helpers import threed

def _(text, disambiguation=None, context='Plotters3D'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class GenericPlotter3D(widget.Widget):
    """Generic plotter."""

    typename = 'genericplotter3d'
    isplotter = True

    @classmethod
    def allowedParentTypes(klass):
        from . import graph3d
        return (graph3d.Graph3D,)

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        widget.Widget.addSettings(s)

        # s.add( setting.Str(
        #     'key', '',
        #     descr = _('Description of the plotted data to appear in key'),
        #     usertext=_('Key text')) )
        s.add( setting.Axis(
            'xAxis', 'x', 'x',
            descr=_('Name of X-axis to use'),
            usertext=_('X axis')) )
        s.add( setting.Axis(
            'yAxis', 'y', 'y',
            descr=_('Name of Y-axis to use'),
            usertext=_('Y axis')) )
        s.add( setting.Axis(
            'zAxis', 'z', 'z',
            descr=_('Name of Z-axis to use'),
            usertext=_('Z axis')) )

    def autoColor(self, painter, dataindex=0):
        """Automatic color for plotting."""
        return painter.docColorAuto(
            painter.helper.autoColorIndex((self, dataindex)))

    def getAxesNames(self):
        """Returns names of axes used."""
        s = self.settings
        return (s.xAxis, s.yAxis, s.zAxis)

    def getNumberKeys(self):
        """Return number of key entries."""
        if self.settings.key:
            return 1
        else:
            return 0

    def getKeyText(self, number):
        """Get key entry."""
        return self.settings.key

    def drawKeySymbol(self, number, painter, x, y, width, height):
        """Draw the plot symbol and/or line at (x,y) in a box width*height.

        This is used to plot a key
        """
        pass

    def getAxisLabels(self, direction):
        """Get labels for datapoints and coordinates, or None if none.
        direction is 'horizontal' or 'vertical'

        return (labels, coordinates)
        """
        return (None, None, None)

    def fetchAxes(self):
        """Returns the axes for this widget"""

        axes = self.parent.getAxes(
            (self.settings.xAxis, self.settings.yAxis, self.settings.zAxis))

        # fail if we don't have good axes
        if ( axes[0] is None or axes[0].settings.direction != 'x' or
             axes[1] is None or axes[1].settings.direction != 'y' or
             axes[2] is None or axes[2].settings.direction != 'z' ):
            return None

        return axes

    def fetchAxis(self, var):
        """Return a particular axis given x, y or z"""
        if var == 'x':
            setn = self.settings.xAxis
        elif var == 'y':
            setn = self.settings.yAxis
        elif var == 'z':
            setn = self.settings.zAxis
        return self.parent.getAxes((setn,))[0]

    def lookupAxis(self, axisname):
        """Find widget associated with axisname."""
        w = self.parent
        while w:
            for c in w.children:
                if c.name == axisname and c.isaxis:
                    return c
            w = w.parent
        return None

    def affectsAxisRange(self):
        """Returns information on the following axes.
        format is ( ('x', 'sx'), ('y', 'sy') )
        where key is the axis and value is a provided bound
        """
        return ()

    def requiresAxisRange(self):
        """Requires information about the axis given before providing
        information.
        Format (('sx', 'x'), ('sy', 'y'))
        """
        return ()

    def getRange(self, axis, depname, therange):
        """Update range variable for axis with dependency name given."""
        pass

    def simplifyObjectList(self, objlist):
        """Simplify list of objects, if possible."""
        if len(objlist) == 0:
            return None
        elif len(objlist) == 1:
            return objlist[0]
        else:
            cont = threed.ObjectContainer()
            for o in objlist:
                cont.addObject(o)
            return cont

    def makeClipContainer(self, axes):
        """Make an object container to clip data to axes."""
        return threed.ClipContainer(
            threed.Vec3(
                axes[0].settings.lowerPosition,
                axes[1].settings.lowerPosition,
                axes[2].settings.lowerPosition),
            threed.Vec3(
                axes[0].settings.upperPosition,
                axes[1].settings.upperPosition,
                axes[2].settings.upperPosition))

    def drawToObject(self, painter, painthelper):
        # exit if hidden or function blank
        if self.settings.hide:
            return

        # get axes widgets
        axes = self.fetchAxes()
        if not axes:
            return
        else:
            return self.dataDrawToObject(painter, axes)

    def dataDrawToObject(self, painter, axes):
        """Actually plot the data."""
        pass



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Dialog boxes for tools and dataset plugins."""

import sys

from .. import qtall as qt
from .. import document
from .. import plugins
from .. import utils
from . import exceptiondialog
from .veuszdialog import VeuszDialog

def _(text, disambiguation=None, context="PluginDialog"):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

def handlePlugin(mainwindow, doc, pluginkls):
    """Show plugin dialog or directly execute (if it takes no parameters)."""

    plugin = pluginkls()
    if plugin.has_parameters:
        d = PluginDialog(mainwindow, doc, plugin, pluginkls)
        mainwindow.showDialog(d)
    else:
        fields = {'currentwidget': '/'}
        if mainwindow.treeedit.selwidgets:
            fields = {'currentwidget': mainwindow.treeedit.selwidgets[0].path}
        runPlugin(mainwindow, doc, plugin, fields)

def wordwrap(text, linelength=80):
    """Wrap on a word boundary."""
    out = []
    l = 0
    for w in text.split(' '):
        if w.find('\n') >= 0:
            l = 0
        if l + len(w) > linelength:
            out.append('\n')
            l = 0
        out.append(w)
        l += len(w)
    return ' '.join(out)

class PluginDialog(VeuszDialog):
    """Dialog box class for plugins."""

    def __init__(self, mainwindow, doc, plugininst, pluginkls):
        VeuszDialog.__init__(self, mainwindow, 'plugin.ui')

        reset = self.buttonBox.button(qt.QDialogButtonBox.Reset)
        reset.setAutoDefault(False)
        reset.setDefault(False)
        reset.clicked.connect( self.slotReset)
        self.buttonBox.button(
            qt.QDialogButtonBox.Apply).clicked.connect(self.slotApply)

        self.pluginkls = pluginkls
        self.plugininst = plugininst
        self.document = doc

        title = ': '.join(list(plugininst.menu))
        self.setWindowTitle(title)
        descr = plugininst.description_full
        if plugininst.author:
            descr += '\n ' + _('Author: %s') % plugininst.author
        self.descriptionLabel.setText( wordwrap(descr) )

        self.fieldcntrls = []
        self.fields = []
        self.addFields()

    def addFields(self):
        """Add any fields, removing existing ones if required."""
        layout = self.fieldGroup.layout()

        for line in self.fieldcntrls:
            for cntrl in line:
                layout.removeWidget(cntrl)
                cntrl.deleteLater()
        del self.fieldcntrls[:]

        currentwidget = '/'
        if self.mainwindow.treeedit.selwidgets:
            currentwidget = self.mainwindow.treeedit.selwidgets[0].path
        for row, field in enumerate(self.plugininst.fields):
            if isinstance(field, list) or isinstance(field, tuple):
                for c, f in enumerate(field):
                    cntrls = f.makeControl(self.document, currentwidget)
                    layout.addWidget(cntrls[0], row, c*2)
                    layout.addWidget(cntrls[1], row, c*2+1)
                    self.fieldcntrls.append(cntrls)
                    self.fields.append(f)
            else:
                cntrls = field.makeControl(self.document, currentwidget)
                layout.addWidget(cntrls[0], row, 0)
                layout.addWidget(cntrls[1], row, 1)
                self.fieldcntrls.append(cntrls)
                self.fields.append(field)

    def slotReset(self):
        """Reset fields to defaults."""
        self.addFields()

    def reEditDataset(self, ds, dsname):
        """Open up dataset in dialog for editing."""

        oldfields = ds.pluginmanager.fields
        for field, cntrl in zip(self.fields, self.fieldcntrls):
            field.setControlVal(cntrl, oldfields[field.name])

    def slotApply(self):
        """Use the plugin with the inputted data."""

        # default field
        fields = {'currentwidget': '/'}
        if self.mainwindow.treeedit.selwidgets:
            fields = {
                'currentwidget': self.mainwindow.treeedit.selwidgets[0].path
            }

        # read values from controls
        for field, cntrls in zip(self.fields, self.fieldcntrls):
            fields[field.name] = field.getControlResults(cntrls)

        # run plugin
        plugin = self.pluginkls()
        statustext = runPlugin(self, self.document, plugin, fields)

        # show any results
        self.notifyLabel.setText(statustext)
        qt.QTimer.singleShot(3000, self.notifyLabel.clear)

def runPlugin(window, doc, plugin, fields):
    """Execute a plugin.
    window - parent window
    doc - veusz document
    plugin - plugin object."""

    if isinstance(plugin, plugins.ToolsPlugin):
        mode = 'tools'
    elif isinstance(plugin, plugins.DatasetPlugin):
        mode = 'dataset'
    else:
        raise RuntimeError("Invalid plugin class")

    resultstext = ''
    try:
        with utils.OverrideCursor():
            # use correct operation class for different plugin types
            if mode == 'tools':
                op = document.OperationToolsPlugin(plugin, fields)
            elif mode == 'dataset':
                # a bit of a hack as we don't give currentwidget to this plugin
                del fields['currentwidget']
                op = document.OperationDatasetPlugin(
                    plugin, fields, raiseerrors=True)

            results = doc.applyOperation(op)

            # evaluate datasets using plugin to check it works
            if mode == 'dataset':
                op.validate()
                resultstext = _('Created datasets: ') + ', '.join(results)
            else:
                resultstext = _('Done')

    except (plugins.ToolsPluginException, plugins.DatasetPluginException) as ex:
        # unwind operations
        op.undo(doc)

        qt.QMessageBox.warning(
            window, _("Error in %s") % plugin.name, str(ex))

    except Exception:
        op.undo(doc)

        # show exception dialog
        exceptiondialog.ExceptionDialog(sys.exc_info(), window).exec_()

    return resultstext

def recreateDataset(mainwindow, document, dataset, datasetname):
    """Open dialog to recreate plugin dataset(s)."""

    # make a new instance of the plugin class
    kls = dataset.pluginmanager.plugin.__class__
    newplugin = kls()

    dialog = PluginDialog(mainwindow, document, newplugin, kls)
    mainwindow.showDialog(dialog)
    dialog.reEditDataset(dataset, datasetname)



# data import dialog

#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Module for implementing dialog boxes for importing data in Veusz."""

import os.path
import sys

from .. import qtall as qt
from .. import setting
from .. import utils
from .. import plugins
from . import exceptiondialog
from .veuszdialog import VeuszDialog

def _(text, disambiguation=None, context="ImportDialog"):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class ImportTab(qt.QWidget):
    """Tab for a particular import type."""

    resource = ''
    filetypes = ()       # list of file types handled
    filefilter = None    # name of filter for types for open dialog

    def __init__(self, importdialog, *args):
        """Initialise dialog. importdialog is the import dialog itself."""
        qt.QWidget.__init__(self, *args)
        self.dialog = importdialog
        self.uiloaded = False

    def loadUi(self):
        """Load up UI file."""
        qt.loadUi(os.path.join(
            utils.resourceDirectory, 'ui', self.resource), self)
        self.uiloaded = True

    def reset(self):
        """Reset controls to initial conditions."""
        pass

    def doPreview(self, filename, encoding):
        """Update the preview window, returning whether import
        should be attempted."""
        pass

    def doImport(self, doc, filename, linked, encoding, prefix, suffix, tags):
        """Do the import iteself."""
        pass

    def okToImport(self):
        """Secondary check (after preview) for enabling import button."""
        return True

    def isFiletypeSupported(self, ftype):
        """Is the filetype supported by this tab?"""
        return ftype in self.filetypes

    def useFiletype(self, ftype):
        """If the tab can do something with the selected filetype,
        update itself."""
        pass

importtabs = []
def registerImportTab(name, klass):
    """Register an import tab for the dialog."""
    importtabs.append((name, klass))

class ImportDialog(VeuszDialog):
    """Dialog box for importing data.
    See ImportTab classes above which actually do the work of importing
    """

    dirname = '.'

    def __init__(self, parent, document):

        VeuszDialog.__init__(self, parent, 'import.ui')
        self.document = document

        # whether file import looks likely to work
        self.filepreviewokay = False

        # tabs loaded currently in dialog
        self.tabs = {}
        for tabname, tabclass in importtabs:
            w = tabclass(self)
            self.methodtab.addTab(w, tabname)

        # add promoted plugins
        for p in plugins.importpluginregistry:
            if p.promote_tab is not None:
                from ..dataimport.dialog_plugin import ImportTabPlugins
                w = ImportTabPlugins(self, promote=p.name)
                self.methodtab.addTab(w, p.name)

        self.methodtab.currentChanged.connect(self.slotTabChanged)
        self.browsebutton.clicked.connect(self.slotBrowseClicked)
        self.filenameedit.editTextChanged.connect(self.slotUpdatePreview)

        self.importbutton = self.buttonBox.addButton(
            _("&Import"), qt.QDialogButtonBox.ApplyRole)
        self.importbutton.clicked.connect(self.slotImport)

        self.buttonBox.button(qt.QDialogButtonBox.Reset).clicked.connect(
            self.slotReset)
        self.encodingcombo.currentIndexChanged.connect(self.slotUpdatePreview)

        # add completion for filename
        c = self.filenamecompleter = qt.QCompleter(self)
        self.filenameedit.setCompleter(c)

        # change to tab last used
        self.methodtab.setCurrentIndex(
            setting.settingdb.get('import_lasttab', 0))

        # defaults for prefix and suffix
        self.prefixcombo.default = self.suffixcombo.default = ['', '$FILENAME']

        # default state for check boxes
        self.linkcheckbox.default = True

        # further defaults
        self.encodingcombo.defaultlist = utils.encodings
        self.encodingcombo.defaultval = 'utf_8'

        # load icon for clipboard
        self.clipbutton.setIcon( utils.getIcon('kde-clipboard') )
        qt.QApplication.clipboard().dataChanged.connect(
            self.updateClipPreview)
        self.clipbutton.clicked.connect(self.slotClipButtonClicked)
        self.updateClipPreview()

    def slotBrowseClicked(self):
        """Browse for a data file."""

        fd = qt.QFileDialog(self, _('Browse data file'))
        fd.setFileMode( qt.QFileDialog.ExistingFile )

        # collect filters from tabs
        filters = [_('All files (*)')]
        for i in range(self.methodtab.count()):
            w = self.methodtab.widget(i)
            if w.filefilter:
                ftypes = ' '.join(['*'+t for t in w.filetypes])
                f = '%s (%s)' % (w.filefilter, ftypes)
                filters.append(f)
        fd.setNameFilters(filters)

        lastfilt = setting.settingdb.get('import_filterbrowse')
        if lastfilt in filters:
            fd.selectNameFilter(lastfilt)

        # use filename to guess a path if possible
        filename = self.filenameedit.text()
        if os.path.isdir(filename):
            ImportDialog.dirname = filename
        elif os.path.isdir( os.path.dirname(filename) ):
            ImportDialog.dirname = os.path.dirname(filename)

        fd.setDirectory(ImportDialog.dirname)

        # update filename if changed
        if fd.exec_() == qt.QDialog.Accepted:
            ImportDialog.dirname = fd.directory().absolutePath()
            self.filenameedit.replaceAndAddHistory( fd.selectedFiles()[0] )
            setting.settingdb['import_filterbrowse'] = fd.selectedNameFilter()
            self.guessImportTab()

    def guessImportTab(self):
        """Guess import tab based on filename."""
        filename = self.filenameedit.text()

        ftype = os.path.splitext(filename)[1]
        # strip off any gz, bz2 extensions to get real extension
        while ftype.lower() in ('.gz', '.bz2'):
            ftype = os.path.splitext(filename)[1]
        ftype = ftype.lower()

        # examine from left to right
        # promoted plugins come after plugins
        idx = -1
        for i in range(self.methodtab.count()):
            w = self.methodtab.widget(i)
            if w.isFiletypeSupported(ftype):
                idx = i

        if idx >= 0:
            self.methodtab.setCurrentIndex(idx)
            self.methodtab.widget(idx).useFiletype(ftype)

    def slotUpdatePreview(self, *args):
        """Update preview window when filename or tab changed."""

        # save so we can restore later
        tab = self.methodtab.currentIndex()
        setting.settingdb['import_lasttab'] = tab
        filename = self.filenameedit.text()
        encoding = str(self.encodingcombo.currentText())
        importtab = self.methodtab.currentWidget()

        if encoding == '':
            return

        if isinstance(importtab, ImportTab):
            if not importtab.uiloaded:
                importtab.loadUi()
            self.filepreviewokay = importtab.doPreview(
                filename, encoding)

        # enable or disable import button
        self.enableDisableImport()

    def slotTabChanged(self, tabindex):
        """Change completer depending on tab."""
        self.slotUpdatePreview()
        w = self.methodtab.widget(tabindex)

        if w.filetypes is None:
            filters = ['*.*']
        else:
            filters = ['*'+t for t in w.filetypes]
        model = qt.QDirModel(
            filters, qt.QDir.AllDirs | qt.QDir.Files, qt.QDir.Name)
        self.filenamecompleter.setModel(model)

    def enableDisableImport(self, *args):
        """Disable or enable import button if allowed."""

        importtab = self.methodtab.currentWidget()
        enabled = self.filepreviewokay and importtab.okToImport()

        # actually enable or disable import button
        self.importbutton.setEnabled( enabled )

    def slotImport(self):
        """Do the importing"""

        filename = self.filenameedit.text()
        linked = self.linkcheckbox.isChecked()
        encoding = str(self.encodingcombo.currentText())
        if filename == '{clipboard}':
            linked = False
        else:
            # normalise filename
            filename = os.path.abspath(filename)

        # import according to tab selected
        importtab = self.methodtab.currentWidget()
        prefix, suffix = self.getPrefixSuffix(filename)
        tags = self.tagcombo.currentText().split()

        try:
            with utils.OverrideCursor():
                with self.document.suspend():
                    importtab.doImport(
                        self.document, filename, linked, encoding,
                        prefix, suffix, tags)
        except IOError:
            qt.QMessageBox.warning(
                self, _("Veusz"), _("Could not read file"))
        except Exception:
            # show exception dialog
            d = exceptiondialog.ExceptionDialog(sys.exc_info(), self)
            d.exec_()

    def retnDatasetInfo(self, dsnames, linked, filename):
        """Return a list of information for the dataset names given."""

        lines = [_('Imported data for datasets:')]
        dsnames.sort()
        for name in dsnames:
            ds = self.document.getData(name)
            # build up description
            lines.append(_('%s: %s') % (name, ds.description()))

        # whether the data were linked
        if linked:
            lines.append('')
            lines.append(_('Datasets were linked to file "%s"') % filename)

        return lines

    def getPrefixSuffix(self, filename):
        """Get prefix and suffix values."""
        f = utils.cleanDatasetName( os.path.basename(filename) )
        prefix = self.prefixcombo.lineEdit().text()
        prefix = prefix.replace('$FILENAME', f)
        suffix = self.suffixcombo.lineEdit().text()
        suffix = suffix.replace('$FILENAME', f)
        return prefix, suffix

    def slotReset(self):
        """Reset input fields."""

        self.filenameedit.setText("")
        self.encodingcombo.setCurrentIndex(
            self.encodingcombo.findText("utf_8"))
        self.linkcheckbox.setChecked(True)
        self.prefixcombo.setEditText("")
        self.suffixcombo.setEditText("")

        importtab = self.methodtab.currentWidget()
        importtab.reset()

    def slotClipButtonClicked(self):
        """Clicked clipboard button."""
        self.filenameedit.setText("{clipboard}")

    def updateClipPreview(self):
        """Clipboard contents changed, so update preview if showing clipboard."""

        filename = self.filenameedit.text()
        if filename == '{clipboard}':
            self.slotUpdatePreview()



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Plugins for general operations."""

import random
import re
import fnmatch

import numpy as N

from .. import qtall as qt
from .. import utils
from . import field

def _(text, disambiguation=None, context='ToolsPlugin'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

# add an instance of your class to this list to be registered
toolspluginregistry = []

class ToolsPluginException(RuntimeError):
    """Raise this to report an error doing what was requested.
    """
    pass

class ToolsPlugin:
    # the plugin will get inserted into the menu in a hierarchy based on
    # the elements of this tuple
    menu = (_('Base plugin'),)
    name = 'Base plugin'

    author = ''
    description_short = ''
    description_full = ''

    # if the plugin takes no parameters, set this to False
    has_parameters = True

    def __init__(self):
        """Override this to declare a list of input fields if required."""
        self.fields = []

    def apply(self, commandinterface, fieldresults):
        """Override this option to do the work of the plugin.

        * commandinterface is an instance of the embedding interface,
        which also contains the Root widget node object
        * fieldresults is a dict containing the values of the fields plus
        'currentwidget' which is the path to the current widget

        * Raise an ToolsPluginException(str) to report a problem to the user
        """

#########################################################################

class ColorsRandomize(ToolsPlugin):
    """Randomize the colors used in plotting."""

    menu = (_('Colors'), _('Randomize'))
    name = 'Randomize colors'
    description_short = _('Randomize the colors used in plotting')
    description_full = _('Randomize the colors used in plotting markers, lines or error bars. Random colors in hue, saturation and luminosity (HSV) are chosen between the two colors given.')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "randxy", descr=_("Randomize xy plotters"),
                default=True),
            field.FieldBool(
                "randfunc", descr=_("Randomize function plotters"),
                default=True),
            field.FieldColor(
                'color1', descr=_("Start of color range"),
                default='#404040'),
            field.FieldColor(
                'color2', descr=_("End of color range"),
                default='#ff0004'),
        ]

    def getRandomColor(self, col1, col2):
        """Return RGB name for a random color."""

        H1, H2 = col1.hue(), col2.hue()
        S1, S2 = col1.saturation(), col2.saturation()
        V1, V2 = col1.value(), col2.value()

        def rand(a, b):
            if a > b:
                return random.randint(b, a)
            return random.randint(a, b)

        col = qt.QColor.fromHsv(rand(H1, H2), rand(S1, S2), rand(V1, V2))
        return str(col.name())

    def apply(self, ifc, fields):
        """Do the randomizing."""

        fromwidget = ifc.Root.fromPath(fields['widget'])

        col1 = qt.QColor(fields['color1'])
        col2 = qt.QColor(fields['color2'])

        if fields['randxy']:
            for node in fromwidget.WalkWidgets(widgettype='xy'):
                col = self.getRandomColor(col1, col2)
                node.color.val = col

        if fields['randfunc']:
            for node in fromwidget.WalkWidgets(widgettype='function'):
                node.Line.color.val = self.getRandomColor(col1, col2)

class ColorsSequenceCMap(ToolsPlugin):
    """Color sequence using colormap."""

    menu = (_('Colors'), _('Colormap Sequence'))
    name = 'Create color sequence using color map'
    description_short = _('Make widgets use sequence of colors in a colormap')
    description_full = _(
        'Give new colors to each widget in a sequence using a colormap. '
        'Match can be set to match names of widgets (e.g. "xy*").')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"), default="/"),
            field.FieldBool(
                "colorxy", descr=_("Color xy plotters"), default=True),
            field.FieldBool(
                "colorfunc", descr=_("Color function plotters"), default=True),
            field.FieldText(
                "match", descr=_("Match")),
            field.FieldBool(
                "invert", descr=_("Invert colormap"), default=False),
            field.FieldBool(
                "randomize", descr=_("Randomize order"), default=False),
            field.FieldColormap(
                "colormap", descr=_("Colormap"), default="grey"),
        ]

    def apply(self, ifc, fields):
        """Do the randomizing."""

        fromwidget = ifc.Root.fromPath(fields['widget'])
        match = fields['match'].strip()

        widgets = []
        for w in fromwidget.WalkWidgets():
            if match and not fnmatch.fnmatch(w.name, match):
                continue
            if ( (w.widgettype=='xy' and fields['colorxy']) or
                 (w.widgettype=='function' and fields['colorfunc']) ):
                widgets.append(w)

        # get list of RGBA values
        cvals = ifc.GetColormap(
            fields["colormap"], invert=fields["invert"],
            nvals=max(1, len(widgets)))

        if fields["randomize"]:
            N.random.shuffle(cvals)

        # convert colors to #XXXX format
        for idx, widget in enumerate(widgets):
            if cvals[idx,3] == 255:
                # opaque
                col = "#%02x%02x%02x" % (
                    cvals[idx,0], cvals[idx,1], cvals[idx,2])
            else:
                # with transparency
                col = "#%02x%02x%02x%02x" % (
                    cvals[idx,0], cvals[idx,1], cvals[idx,2], cvals[idx,3])

            t = widget.widgettype
            if t == 'xy':
                widget.color.val = col
            elif t == 'function':
                widget.Line.color.val = col

class ColorsSequence(ToolsPlugin):
    """Color plotters in sequence."""

    menu = (_('Colors'), _('Sequence'))
    name = 'Create color sequence'
    description_short = _('Make widgets use sequence of colors')
    description_full = _('Give new colors to each widget in a sequence between the two colors given.')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "randxy", descr=_("Color xy plotters"),
                default=True),
            field.FieldBool(
                "randfunc", descr=_("Color function plotters"),
                default=True),
            field.FieldColor(
                'color1', descr=_("Start of color range"),
                default='#ff0000'),
            field.FieldColor(
                'color2', descr=_("End of color range"),
                default='#4000ff'),
        ]

    def apply(self, ifc, fields):
        """Do the sequence."""

        fromwidget = ifc.Root.fromPath(fields['widget'])

        col1 = qt.QColor(fields['color1'])
        col2 = qt.QColor(fields['color2'])
        H1, H2 = col1.hue(), col2.hue()
        S1, S2 = col1.saturation(), col2.saturation()
        V1, V2 = col1.value(), col2.value()

        # add up total number of widgets
        numwidgets = (
            len( list(fromwidget.WalkWidgets(widgettype='xy')) ) +
            len( list(fromwidget.WalkWidgets(widgettype='function')) ) )

        def colatidx(i):
            """Get color in range 0...numwidgets-1."""
            div = max(numwidgets-1, 1)

            H = i * (H2-H1) / div + H1
            S = i * (S2-S1) / div + S1
            V = i * (V2-V1) / div + V1
            return str(qt.QColor.fromHsv(H, S, V).name())

        idx = 0
        for node in fromwidget.WalkWidgets():
            t = node.widgettype
            if fields['randxy'] and t == 'xy':
                col = colatidx(idx)
                idx += 1
                node.color.val = col

            if fields['randfunc'] and t == 'function':
                node.Line.color.val = colatidx(idx)
                idx += 1

class ColorsReplace(ToolsPlugin):
    """Replace one color by another."""

    menu = (_('Colors'), _('Replace'))
    name = 'Replace colors'
    description_short = _('Search and replace colors')
    description_full = _('Searches for a color and replaces it with a different color')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "follow", descr=_("Change references and defaults"),
                default=True),
            field.FieldColor(
                'color1', descr=_("Color to change"),
                default='black'),
            field.FieldColor(
                'color2', descr=_("Replacement color"),
                default='red'),
        ]

    def apply(self, ifc, fields):
        """Do the color search and replace."""

        fromcol = qt.QColor(fields['color1'])

        def walkNodes(node):
            """Walk nodes, changing values."""
            if node.type == 'setting' and node.settingtype == 'color':
                # only follow references if requested
                if node.isreference:
                    if fields['follow']:
                        node = node.resolveReference()
                    else:
                        return

                # evaluate into qcolor to make sure is a true match
                if qt.QColor(node.val) == fromcol:
                    node.val = fields['color2']
            else:
                for c in node.children:
                    walkNodes(c)

        fromwidget = ifc.Root.fromPath(fields['widget'])
        walkNodes(fromwidget)

class ColorsSwap(ToolsPlugin):
    """Swap colors used in plotting."""

    menu = (_('Colors'), _('Swap'))
    name = 'Swap colors'
    description_short = _('Swap two colors')
    description_full = _('Swaps two colors in the plot')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "follow", descr=_("Change references and defaults"),
                default=True),
            field.FieldColor(
                'color1', descr=_("First color"),
                default='black'),
            field.FieldColor(
                'color2', descr=_("Second color"),
                default='red'),
        ]

    def apply(self, ifc, fields):
        """Do the color search and replace."""

        col1 = qt.QColor(fields['color1'])
        col2 = qt.QColor(fields['color2'])

        def walkNodes(node):
            """Walk nodes, changing values."""
            if node.type == 'setting' and node.settingtype == 'color':
                # only follow references if requested
                if node.isreference:
                    if fields['follow']:
                        node = node.resolveReference()
                    else:
                        return

                # evaluate into qcolor to make sure is a true match
                if qt.QColor(node.val) == col1:
                    node.val = fields['color2']
                elif qt.QColor(node.val) == col2:
                    node.val = fields['color1']
            else:
                for c in node.children:
                    walkNodes(c)

        fromwidget = ifc.Root.fromPath(fields['widget'])
        walkNodes(fromwidget)

class TextReplace(ToolsPlugin):
    """Randomize the colors used in plotting."""

    menu = (_('General'), _('Replace text'))
    name = 'Replace text'
    description_short = _('Search and replace text in settings')
    description_full = _('Searches for text in a setting and replaces it')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "follow", descr=_("Change references and defaults"),
                default=True),
            field.FieldBool(
                "onlystr", descr=_("Change only textual data"),
                default=False),
            field.FieldText(
                'text1', descr=_("Text to change"),
                default=''),
            field.FieldText(
                'text2', descr=_("Replacement text"),
                default=''),
        ]

    def apply(self, ifc, fields):
        """Do the search and replace."""

        def walkNodes(node):
            """Walk nodes, changing values."""
            if node.type == 'setting':
                # only follow references if requested
                if node.isreference:
                    if fields['follow']:
                        node = node.resolveReference()
                    else:
                        return

                val = node.val
                # try to change if a string, and not only strings or type is string
                if isinstance(val, str) and (
                        not fields['onlystr'] or node.settingtype == 'str'):
                    # update text if it changes
                    val2 = val.replace(fields['text1'], fields['text2'])
                    if val != val2:
                        try:
                            node.val = val2
                        except utils.InvalidType:
                            pass
            else:
                for c in node.children:
                    walkNodes(c)

        fromwidget = ifc.Root.fromPath(fields['widget'])
        walkNodes(fromwidget)

class WidgetsClone(ToolsPlugin):
    """Take a widget and children and clone them."""

    menu = (_('Widgets'), _('Clone for datasets'))
    name = 'Clone widgets for datasets'
    description_short = _('Clones a widget and its children for datasets')
    description_full = _('Take a widget and its children and clone it, plotting different sets of data in each clone.\nHint: Use a "*" in the name of a replacement dataset to match multiple datasets, e.g. x_*')

    def __init__(self):
        """Construct plugin."""
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Clone widget"),
                default=""),
            field.FieldDataset(
                'ds1', descr=_("Dataset 1 to change"),
                default=''),
            field.FieldDatasetMulti(
                'ds1repl',
                descr=_("Replacement(s) for dataset 1")),
            field.FieldDataset(
                'ds2', descr=_("Dataset 2 to change (optional)"),
                default=''),
            field.FieldDatasetMulti(
                'ds2repl',
                descr=_("Replacement(s) for dataset 2")),
            field.FieldBool(
                "names", descr=_("Build new names from datasets"),
                default=True),
        ]

    def apply(self, ifc, fields):
        """Do the cloning."""

        def expanddatasets(dslist):
            """Expand * and ? in dataset names."""
            datasets = []
            for ds in dslist:
                if ds.find('*') == -1 and ds.find('?') == -1:
                    datasets.append(ds)
                else:
                    dlist = fnmatch.filter(ifc.GetDatasets(), ds)
                    dlist.sort()
                    datasets += dlist
            return datasets

        def chainpairs(dslist1, dslist2):
            """Return pairs of datasets, repeating if necessary."""
            if not dslist1:
                dslist1 = ['']
            if not dslist2:
                dslist2 = ['']

            end1 = end2 = False
            idx1 = idx2 = 0
            while True:
                if idx1 >= len(ds1repl):
                    idx1 = 0
                    end1 = True
                if idx2 >= len(ds2repl):
                    idx2 = 0
                    end2 = True
                if end1 and end2:
                    break

                yield dslist1[idx1], dslist2[idx2]
                idx1 += 1
                idx2 += 1

        def walkNodes(node, dsname, dsrepl):
            """Walk nodes, changing datasets."""
            if node.type == 'setting':
                if node.settingtype in (
                        'dataset', 'dataset-extended',
                        'dataset-or-floatlist', 'dataset-or-str'):
                    # handle single datasets
                    if node.val == dsname:
                        node.val = dsrepl
                elif node.settingtype == 'dataset-multi':
                    # settings with multiple datasets
                    out = list(node.val)
                    for i, v in enumerate(out):
                        if v == dsname:
                            out[i] = dsrepl
                    if tuple(out) != node.val:
                        node.val = out
            else:
                for c in node.children:
                    walkNodes(c, dsname, dsrepl)

        # get names of replacement datasets
        ds1repl = expanddatasets(fields['ds1repl'])
        ds2repl = expanddatasets(fields['ds2repl'])

        # make copies of widget and children for each pair of datasets
        widget = ifc.Root.fromPath(fields['widget'])
        for ds1r, ds2r in chainpairs(ds1repl, ds2repl):
            # construct a name
            newname = None
            if fields['names']:
                newname = widget.name
                if ds1r:
                    # / cannot be in dataset name
                    flt1 = ds1r.replace('/', '_')
                    newname += ' ' + flt1
                if ds2r:
                    flt2 = ds2r.replace('/', '_')
                    newname += ' ' + flt2

            # make the new widget (and children)
            newwidget = widget.Clone(widget.parent, newname=newname)

            # do replacement of datasets
            if fields['ds1']:
                walkNodes(newwidget, fields['ds1'], ds1r)
            if fields['ds2']:
                walkNodes(newwidget, fields['ds2'], ds2r)

class FontSize(ToolsPlugin):
    """Increase or decrease the font size."""

    def __init__(self, dirn):
        """Construct plugin.
        dirn == 1: increase sizes
        dirn == -1: decrease sizes
        """
        self.dirn = dirn
        self.fields = [
            field.FieldWidget(
                "widget", descr=_("Start from widget"),
                default="/"),
            field.FieldBool(
                "follow", descr=_("Change references and defaults"),
                default=True),
            field.FieldFloat(
                "delta", descr=_("Change by value"),
                default=2),
        ]

    def apply(self, ifc, fields):
        """Do the search and replace."""

        pt_re = re.compile(r'^([\d.]+)[ ]*pt$')
        delta = fields['delta']
        changed = set()

        def walkNodes(node):
            """Walk nodes, changing values."""
            if node.type == 'setting':
                if node.name == 'size':
                    # find size setting with sibling font (improve this)
                    if not hasattr(node.parent, 'font'):
                        return

                    # only follow references if requested
                    if node.isreference:
                        if fields['follow']:
                            node = node.resolveReference()
                        else:
                            return

                    # avoid doing things more than once
                    p = node.path
                    if p in changed:
                        return
                    changed.add(p)

                    # change point size if requested
                    m = pt_re.match(node.val)
                    if m:
                        pt = float(m.group(1)) + delta*self.dirn
                        if pt < 0: pt = 0.1
                        node.val = '%gpt' % pt
            else:
                for c in node.children:
                    walkNodes(c)

        fromwidget = ifc.Root.fromPath(fields['widget'])
        walkNodes(fromwidget)

class FontSizeIncrease(FontSize):
    menu = (_('General'), _('Increase font sizes'))
    name = 'Increase font sizes'
    description_short = _('Increase font sizes')
    description_full = _('Increase font sizes by number of points given')

    def __init__(self):
        FontSize.__init__(self, 1)

class FontSizeDecrease(FontSize):
    menu = (_('General'), _('Decrease font sizes'))
    name = 'Decrease font sizes'
    description_short = _('Decrease font sizes')
    description_full = _('Decrease font sizes by number of points given')

    def __init__(self):
        FontSize.__init__(self, -1)

toolspluginregistry += [
    ColorsRandomize,
    ColorsSequence,
    ColorsSequenceCMap,
    ColorsReplace,
    ColorsSwap,
    TextReplace,
    WidgetsClone,
    FontSizeIncrease,
    FontSizeDecrease,
]



#    Copyright (C) 2018 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import math

from .. import qtall as qt
from .. import document
from .. import setting
from . import widget
from . import controlgraph
from ..helpers import threed

def _(text, disambiguation=None, context='Scene3D'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

# define 2nd and 3rd lighting classes
class Lighting3D_2(setting.Lighting3D):
    def __init__(self, name, **args):
        setting.Lighting3D.__init__(self, name, **args)
        self.get('enable').newDefault(False)
        self.get('color').newDefault('red')
        self.get('x').newDefault(2)
class Lighting3D_3(setting.Lighting3D):
    def __init__(self, name, **args):
        setting.Lighting3D.__init__(self, name, **args)
        self.get('enable').newDefault(False)
        self.get('color').newDefault('blue')
        self.get('x').newDefault(-2)

class Scene3D(widget.Widget):
    """3D scene containing other widgets."""

    typename='scene3d'
    allowusercreation = True
    description = _('3d scene')

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        widget.Widget.addSettings(s)

        s.add( setting.FloatSlider(
            'xRotation',
            0,
            minval=-180, maxval=180, step=15, tick=45,
            descr=_('Rotation around x axis (°)'),
            usertext=_('X rotation') ))
        s.add( setting.FloatSlider(
            'yRotation',
            35.,
            minval=-180, maxval=180, step=15, tick=45,
            descr=_('Rotation around y axis (°)'),
            usertext=_('Y rotation') ))
        s.add( setting.FloatSlider(
            'zRotation',
            0,
            minval=-180, maxval=180, step=15, tick=45,
            descr=_('Rotation around z axis (°)'),
            usertext=_('Z rotation') ))

        s.add( setting.FloatSlider(
            'distance',
            5,
            minval=1, maxval=50, step=0.5, tick=5, scale=0.1,
            descr=_('Viewing distance'),
            usertext=_('Distance') ))

        s.add( setting.FloatOrAuto(
            'size',
            'Auto',
            minval=0,
            descr=_('Automatic or fixed graph size scaling value'),
            usertext=_('Size'),
            formatting=True ))

        s.add( setting.Choice(
            'rendermode',
            ('painters', 'bsp'),
            'painters',
            uilist=("Fast (Painter's)",
                    "Accurate (BSP)"),
            usertext=_('Render method'),
            descr=_('Method used to draw 3D plot') ))

        s.add( setting.Distance(
            'leftMargin',
            '1cm',
            descr=_('Distance from left of graph to edge'),
            usertext=_('Left margin'),
            formatting=True) )
        s.add( setting.Distance(
            'rightMargin',
            '1cm',
            descr=_('Distance from right of graph to edge'),
            usertext=_('Right margin'),
            formatting=True) )
        s.add( setting.Distance(
            'topMargin',
            '1cm',
            descr=_('Distance from top of graph to edge'),
            usertext=_('Top margin'),
            formatting=True) )
        s.add( setting.Distance(
            'bottomMargin',
            '1cm',
            descr=_('Distance from bottom of graph to edge'),
            usertext=_('Bottom margin'),
            formatting=True) )

        s.add(setting.Lighting3D(
            'Lighting1',
            descr=_('Lighting (1)'),
            usertext=_('Lighting (1)')),
            pixmap='settings_lighting' )
        s.add(Lighting3D_2(
            'Lighting2',
            descr=_('Lighting (2)'),
            usertext=_('Lighting (2)')),
            pixmap='settings_lighting' )
        s.add(Lighting3D_3(
            'Lighting3',
            descr=_('Lighting (3)'),
            usertext=_('Lighting (3)')),
            pixmap='settings_lighting' )

    @classmethod
    def allowedParentTypes(self):
        from . import page, grid
        return (page.Page, grid.Grid)

    def getMargins(self, painthelper):
        """Use settings to compute margins."""
        s = self.settings
        return (
            s.get('leftMargin').convert(painthelper),
            s.get('topMargin').convert(painthelper),
            s.get('rightMargin').convert(painthelper),
            s.get('bottomMargin').convert(painthelper)
        )

    def makeObjects(self, painter, bounds, painthelper):
        """Make objects, returning root"""

        s = self.settings

        # do no painting if hidden
        if s.hide:
            return

        root = threed.ObjectContainer()
        root.objM = threed.rotate3M4(
            s.xRotation/180.*math.pi,
            s.yRotation/180.*math.pi,
            s.zRotation/180.*math.pi)

        # build 3d scene from children
        for c in self.children:
           obj = c.drawToObject(painter, painthelper)
           if obj:
               root.addObject(obj)

        return root

    def makeScene(self, painter):
        """Make Scene and Camera objects."""

        s = self.settings

        camera = threed.Camera()
        # camera necessary to make x,y,z coordinates point in the
        # right direction, with the origin in the lower left towards
        # the viewer
        camera.setPointing(
            threed.Vec3(0,  0, -s.distance),
            threed.Vec3(0,  0,  0),
            threed.Vec3(0, -1,  0))
        camera.setPerspective(90, 1, 100)

        mode = {
            'painters': threed.Scene.RENDER_PAINTERS,
            'bsp': threed.Scene.RENDER_BSP,
        }[s.rendermode]
        scene = threed.Scene(mode)

        # add lighting if enabled
        for light in s.Lighting1, s.Lighting2, s.Lighting3:
            if light.enable:
                scene.addLight(
                    # FIXME: z and y negative here to make direction
                    # correct relative to camera origin
                    threed.Vec3(light.x, -light.y, -light.z),
                    light.get('color').color(painter),
                    light.intensity*0.01)

        return scene, camera

    def draw(self, parentposn, painthelper, outerbounds=None):
        '''Update the margins before drawing.'''

        bounds = self.computeBounds(parentposn, painthelper)

        painter = painthelper.painter(self, bounds)
        painthelper.setControlGraph(self, [
            controlgraph.ControlMarginBox(
                self, bounds, outerbounds, painthelper)])

        root = self.makeObjects(painter, bounds, painthelper)
        if root is None:
            return bounds

        scene, camera = self.makeScene(painter)

        # def ptToScreen(pt):
        #     pt_v = (camera.viewM*root.objM)*threed.Vec4(pt[0],pt[1],pt[2],1)
        #     pt_proj = threed.calcProjVec(camera.perspM, pt_v)
        #     pt_screen = threed.projVecToScreen(scene.screenM, pt_proj)
        #     return pt_screen,qt.QPointF(pt_screen.get(0), pt_screen.get(1))

        # finally render the scene
        scale = self.settings.size
        if scale == 'Auto':
            scale = -1
        with painter:
            scene.render(
                root,
                painter, camera,
                bounds[0], bounds[1], bounds[2], bounds[3], scale)

        #     painter.setPen(qt.QPen(qt.Qt.red))
        #     origin = ptToScreen((0,0,0))[1]
        #     for axpt in ((0.5,0,0),(0,0.5,0),(0,0,0.5)):
        #         painter.drawLine(origin, ptToScreen(axpt)[1])

        # axx = threed.Vec4(0,0.5,0,1)
        # threed.solveInverseRotation(camera.viewM, camera.perspM, scene.screenM, axx, ptToScreen((0,0.5,0))[0])

    def identifyWidgetAtPoint(self, painthelper, bounds, scaling, x, y):
        painter = document.PainterRoot()
        painter.updateMetaData(painthelper)

        root = self.makeObjects(painter, bounds, painthelper)
        if root is None:
            return self
        scene, camera = self.makeScene(painter)

        sizescale = self.settings.size
        if sizescale == 'Auto':
            sizescale = -1
        widgetid = scene.idPixel(
            root, painter, camera,
            bounds[0], bounds[1], bounds[2], bounds[3], sizescale,
            scaling, int(x), int(y))

        # recursive check id of children against returned value
        widget = [self]
        def checkwidget(r):
            for c in r.children:
                if id(c) == widgetid:
                    widget[0] = c
                checkwidget(c)

        checkwidget(self)
        return widget[0]

    def updateControlItem(self, cgi):
        """Area moved or resized - call helper routine to move self."""
        cgi.setWidgetMargins()

document.thefactory.register(Scene3D)



#    pickable.py
#    stuff related to the Picker (aka Read Data) tool

#    Copyright (C) 2011 Benjamin K. Stuhl
#    Email: Benjamin K. Stuhl <bks24@cornell.edu>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

import numpy as N

class PickInfo:
    """Encapsulates the results of a Pick operation. graphpos and coords are
       numeric (x,y) tuples, labels are the textual labels for the x and y
       datasets, and index is some object that the picker can use to figure out
       what the 'next' and 'previous' points are. index must implement __str__();
       return '' if it has no user-visible meaning."""
    def __init__(self, widget=None, graphpos=None, labels=None, coords=None, index=None):
        self.widget = widget
        self.graphpos = graphpos
        self.labels = labels
        self.coords = coords
        self.index = index
        self.distance = float('inf')
        self.displaytype = ('numeric', 'numeric')

    def __bool__(self):
        return bool(
            self.widget and self.graphpos and self.labels and self.coords)

class Index:
    """A class containing all the state a GenericPickable needs to find the
       next or previous point"""
    def __init__(self, ivar, index, sign):
        self.ivar = ivar
        self.index = index
        self.sign = sign

        # default to not trusting the actual index to be meaningful
        self.useindex = False

    def __str__(self):
        if not self.useindex:
            return ''
        else:
            # 1-based index
            return str(self.index+1)

def _chooseOrderingSign(m, c, p):
    """Figures out whether p or m is visually right of c"""
    assert c is not None

    if p is not None and m is not None:
        if p[0] > m[0] or (p[0] == m[0] and p[1] < m[1]):
            # p is visually to the right of or above m
            return 1
        else:
            return -1
    elif p is not None:
        if p[0] > c[0]:
            # p is visually right of c
            return 1
        else:
            return -1
    elif m is not None:
        if m[0] < c[0]:
            # m is visually left of c
            return 1
        else:
            return -1
    else:
        assert m is not None or p is not None

class GenericPickable:
    """Utility class which abstracts the math of picking the closest point out
       of a list of points"""

    def __init__(self, widget, labels, vals, graphvals):
        self.widget = widget
        self.labels = labels
        self.xvals, self.yvals = vals
        self.xgraph, self.ygraph = graphvals

    def _pickSign(self, i):
        if len(self.xgraph) <= 1:
            # we only have one element, so it doesn't matter anyways
            return 1

        # go backwards to get previous finite point
        mi = i-1
        while mi >= 0 and not N.isfinite(self.xgraph[mi]+self.ygraph[mi]):
            mi -= 1
        if mi < 0:
            m = None
        else:
            m = self.xgraph[mi], self.ygraph[mi]

        # point in centre
        c = self.xgraph[i], self.ygraph[i]

        # find next finite point
        pi = i+1
        while pi < len(self.xgraph) and not N.isfinite(self.xgraph[pi]+self.ygraph[pi]):
            pi += 1
        if pi == len(self.xgraph):
            p = None
        else:
            p = self.xgraph[pi], self.ygraph[pi]

        if p is None and m is None:
            # only 1 point after removing non-finite points
            return 1

        return _chooseOrderingSign(m, c, p)

    def pickPoint(self, x0, y0, bounds, distance_direction):
        info = PickInfo(self.widget, labels=self.labels)

        if self.widget.settings.hide:
            return info

        if self.xvals is None or self.yvals is None:
            return info
        if len(self.xgraph) == 0 or len(self.ygraph) == 0:
            return info

        # calculate distances
        if distance_direction == 'vertical':
            # measure distance along y
            dist = N.abs(self.ygraph - y0)
        elif distance_direction == 'horizontal':
            # measure distance along x
            dist = N.abs(self.xgraph - x0)
        elif distance_direction == 'radial':
            # measure radial distance
            dist = N.sqrt((self.xgraph - x0)**2 + (self.ygraph - y0)**2)
        else:
            # programming error
            assert (distance_direction == 'radial' or
                    distance_direction == 'vertical' or
                    distance_direction == 'horizontal')

        # ignore points which are offgraph or not finite
        with N.errstate(invalid='ignore'):
            outofbounds = (
                (self.xgraph < bounds[0]) | (self.xgraph > bounds[2]) |
                (self.ygraph < bounds[1]) | (self.ygraph > bounds[3]) |
                ~N.isfinite(self.xgraph) | ~N.isfinite(self.ygraph) )
        dist[outofbounds] = N.inf

        m = N.min(dist)
        # if there are multiple equidistant points, arbitrarily take
        # the first one

        try:
            i = N.nonzero(dist == m)[0][0]
        except IndexError:
            return info

        info.graphpos = self.xgraph[i], self.ygraph[i]
        info.coords = self.xvals[i], self.yvals[i]
        info.distance = m
        info.index = Index(self.xvals[i], i, self._pickSign(i))

        return info

    def pickIndex(self, oldindex, direction, bounds):
        info = PickInfo(self.widget, labels=self.labels)

        if self.widget.settings.hide:
            return info

        if self.xvals is None or self.yvals is None:
            return info

        if oldindex.index is None:
            # no explicit index, so find the closest location to the previous
            # independent variable value
            i = N.logical_not( N.logical_or(
                    self.xvals < oldindex.ivar, self.xvals > oldindex.ivar) )

            # and pick the next
            if oldindex.sign == 1:
                i = max(N.nonzero(i)[0])
            else:
                i = min(N.nonzero(i)[0])
        else:
            i = oldindex.index

        if direction == 'right':
            incr = oldindex.sign
        elif direction == 'left':
            incr = -oldindex.sign
        else:
            assert direction == 'right' or direction == 'left'

        i += incr

        # skip points that are outside of the bounds or are not finite
        while (i >= 0 and i < len(self.xgraph) and
                ( not N.isfinite(self.xgraph[i]) or
                  not N.isfinite(self.ygraph[i]) or
                  (self.xgraph[i] < bounds[0] or self.xgraph[i] > bounds[2] or
                   self.ygraph[i] < bounds[1] or self.ygraph[i] > bounds[3]) )):
            i += incr

        if i < 0 or i >= len(self.xgraph):
            return info

        info.graphpos = self.xgraph[i], self.ygraph[i]
        info.coords = self.xvals[i], self.yvals[i]
        info.index = Index(self.xvals[i], i, oldindex.sign)

        return info

class DiscretePickable(GenericPickable):
    """A specialization of GenericPickable that knows how to deal with widgets
       with axes and data sets"""
    def __init__(self, widget, xdata_propname, ydata_propname, mapdata_fn):
        s = widget.settings
        doc = widget.document
        self.xdata = xdata = s.get(xdata_propname).getData(doc)
        self.ydata = ydata = s.get(ydata_propname).getData(doc)
        labels = s.__getattr__(xdata_propname), s.__getattr__(ydata_propname)

        if not xdata or not ydata or not mapdata_fn:
            GenericPickable.__init__( self, widget, labels, (None, None), (None, None) )
            return

        # take data, ensure same size and map it
        x, y = xdata.data, ydata.data
        minlen = min(len(x), len(y))
        x, y = x[:minlen], y[:minlen]
        xs, ys = mapdata_fn(x, y)

        # and set us up with the mapped data
        GenericPickable.__init__( self, widget, labels, (x, y), (xs, ys) )

    def pickPoint(self, x0, y0, bounds, distance_direction):
        info = GenericPickable.pickPoint(self, x0, y0, bounds, distance_direction)
        if not info:
            return None

        info.displaytype = (self.xdata.displaytype, self.ydata.displaytype)

        # indicies are persistent
        info.index.useindex = True
        return info

    def pickIndex(self, oldindex, direction, bounds):
        info = GenericPickable.pickIndex(self, oldindex, direction, bounds)
        info.displaytype = (self.xdata.displaytype, self.ydata.displaytype)

        if not info:
            return info

        # indicies are persistent
        info.index.useindex = True
        return info



#    Copyright (C) 2013 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

'''An axis based on a function of another axis.'''

import numpy as N

from .. import qtall as qt
from .. import setting
from .. import document

from . import axis

def _(text, disambiguation=None, context='FunctionAxis'):
    '''Translate text.'''
    return qt.QCoreApplication.translate(context, text, disambiguation)

class AxisError(RuntimeError):
    pass

class FunctionError(AxisError):
    pass

def solveFunction(function, vals, mint=None, maxt=None):
    '''Solve a function for a list of values (vals), if we don't know
    where the solution lies. function is a function to call.

    This tries a range of possible input values, and uses binary
    search to refine the solution.

    mint and maxt are the bounds to use when solving
    '''

    xvals = N.array((
        -1e90, -1e70, -1e50, -1e40, -1e30, -1e20,
        -1e10, -1e8, -1e6, -1e5, -1e4,
        -1e3, -1e2, -1e1, -4e0, -2e0, -1e0,
        -1e-1, -1e-2, -1e-3, -1e-4,
        -1e-6, -1e-8, -1e-10, -1e-12, -1e-14,
        -1e-18, -1e-22, -1e-26, -1e-30, -1e-34,
        -1e-40, -1e-50, -1e-70, -1e-90,
        0,
        1e-90, 1e-70, 1e-50, 1e-40,
        1e-34, 1e-30, 1e-26, 1e-22, 1e-18,
        1e-14, 1e-12, 1e-10, 1e-8, 1e-6,
        1e-4, 1e-3, 1e-2, 1e-1,
        1e0, 2e0, 4e0, 1e1, 1e2, 1e3,
        1e4, 1e5, 1e6, 1e8, 1e10,
        1e20, 1e30, 1e40, 1e50, 1e70, 1e90
    ))

    if mint is not None:
        xvals = N.hstack(( mint, xvals[xvals > mint] ))
    if maxt is not None:
        xvals = N.hstack(( xvals[xvals < maxt], maxt ))

    # yvalue in correct shape
    try:
        yvals = function(xvals) + N.zeros(len(xvals))
    except Exception as e:
        raise FunctionError(_('Error evaluating function: %s') % str(e))

    anynan = N.any( N.isnan(yvals) )
    if anynan:
        raise FunctionError(
            _('Invalid regions in function (try setting minimum or maximum t)')
        )

    # remove any infinite regions
    f = N.isfinite(yvals)
    xfilt = xvals[f]
    yfilt = yvals[f]

    if len(yfilt) < 2:
        raise FunctionError(_('Solutions to equation cannot be found'))

    # check for monotonicity
    delta = yfilt[1:] - yfilt[:-1]
    pos, neg = N.all(delta >= 0), N.all(delta <= 0)
    if not (pos or neg):
        raise FunctionError(
            _('Not a monotonic function (try setting minimum or maximum t)'))
    if pos and neg:
        raise FunctionError(_('Constant function'))

    # easier if the values are increasing only
    if neg:
        yfilt = yfilt[::-1]
        xfilt = xfilt[::-1]

    # do binary search for each input value
    out = []
    for thisval in vals:
        # renorm to zero
        ydelta = yfilt - thisval

        # solution is between this and the next
        idx = N.searchsorted(ydelta, 0.)
        if idx == 0:
            if ydelta[0] == 0.:
                # work around value being at start of array
                idx = 1
            else:
                raise AxisError(_('No solution found'))
        elif idx == len(ydelta):
            raise AxisError(_('No solution found'))

        x1, x2 = xfilt[idx-1], xfilt[idx]
        y1, y2 = ydelta[idx-1], ydelta[idx]

        # binary search
        tol = abs(1e-6 * thisval)
        for i in range(30):
            # print x1, y1, "->", x2, y2

            if abs(y1) <= tol and abs(y1) < abs(y2):
                x2, y2 = x1, y1
                break   # found solution
            if abs(y2) <= tol:
                x1, y1 = x2, y2
                break   # found solution

            if y1 == y2 or ((y1<0) and (y2<0)) or ((y1>0) and (y2>0)):
                raise AxisError(_('No solution found'))

            ### This is a bit faster, but bisection is simpler
            # xv = N.linspace(x1, x2, num=100)
            # yv = function(xv) + xv*0. - thisval

            # idx = N.searchsorted(yv, 0.)
            # if idx == 0:
            #     idx = 1

            # x1 = xv[idx-1]
            # y1 = yv[idx-1]
            # x2 = xv[idx]
            # y2 = yv[idx]

            x3 = 0.5*(x1+x2)
            y3 = function(x3) - thisval
            if not N.isfinite(y3):
                raise AxisError(_('Non-finite value encountered'))

            if y3 < 0:
                x1 = x3
                y1 = y3
            else:
                x2 = x3
                y2 = y3

        out.append(0.5*(x1+x2))

    return out

class AxisFunction(axis.Axis):
    '''An axis using an function of another axis.'''

    typename = 'axis-function'
    description = 'An axis based on a function of the values of another axis'

    def __init__(self, *args, **argsv):
        axis.Axis.__init__(self, *args, **argsv)

        self.cachedfuncobj = None
        self.cachedbounds = None
        self.funcchangeset = -1
        self.boundschangeset = -1

    @classmethod
    def addSettings(klass, s):
        '''Construct list of settings.'''
        axis.Axis.addSettings(s)

        s.add( setting.BoolSwitch(
            'linked', False,
            settingsfalse=('min', 'max'),
            settingstrue=('linkedaxis',),
            descr=_('Link axis to another axis'),
            usertext=_('Linked') ), 0 )

        s.add( setting.Str(
            'function', 't',
            descr=_('Monotonic function (use t as variable)'),
            usertext=_('Function')), 1 )
        s.add( setting.Axis(
            'linkedaxis', '', 'both',
            descr=_('Axis which this axis is based on'),
            usertext=_('Linked axis')), 6 )
        s.add( setting.FloatOrAuto(
            'mint', 'Auto',
            descr=_('Minimum value of t or Auto'),
            usertext=('Min t')), 7 )
        s.add( setting.FloatOrAuto(
            'maxt', 'Auto',
            descr=_('Maximum value of t or Auto'),
            usertext=('Max t')), 8 )

        s.get('autoRange').hidden = True
        s.get('autoRange').newDefault('exact')

    @property
    def userdescription(self):
        """User friendly description."""
        s = self.settings
        return _("axis='%s', function='%s'") % (s.linkedaxis, s.function)

    def logError(self, ex):
        '''Write error message to document log for exception ex.'''
        self.document.log(
            _("Error in axis-function (%s): '%s'") % (
                self.settings.function, str(ex)))

    def getMinMaxT(self):
        '''Get minimum and maximum t.'''
        mint = self.settings.mint
        if mint == 'Auto':
            mint = None
        maxt = self.settings.maxt
        if maxt == 'Auto':
            maxt = None
        return mint, maxt

    def getFunction(self):
        '''Check whether function needs to be compiled.'''

        if self.funcchangeset == self.document.changeset:
            return self.cachedfuncobj
        self.funcchangeset = self.document.changeset

        compiled = self.document.evaluate.compileCheckedExpression(
            self.settings.function.strip())

        if compiled is None:
            self.cachedfuncobj = None
        else:
            # a python function for doing the evaluation and handling
            # errors
            env = self.document.evaluate.context.copy()

            def function(t):
                env['t'] = t
                try:
                    return eval(compiled, env)
                except Exception as e:
                    self.logError(e)
                    return N.nan + t
            self.cachedfuncobj = function

            mint, maxt = self.getMinMaxT()
            try:
                solveFunction(function, [0.], mint=mint, maxt=maxt)
            except FunctionError as e:
                self.logError(e)
                self.cachedfuncobj = None
            except AxisError:
                pass

        return self.cachedfuncobj

    def invertFunctionVals(self, vals):
        '''Convert values which are a function of fn and compute t.'''
        fn = self.getFunction()
        if fn is None:
            return None
        mint, maxt = self.getMinMaxT()
        try:
            return solveFunction(fn, vals, mint=mint, maxt=maxt)
        except Exception as e:
            self.logError(e)
            return None

    def lookupAxis(self, axisname):
        '''Find widget associated with axisname.'''
        w = self.parent
        while w:
            for c in w.children:
                if ( c.name == axisname and c.isaxis and
                     c is not self ):
                    return c
            w = w.parent
        return None

    def isLinked(self):
        '''Is this axis linked to another?'''
        return self.settings.linked

    def getLinkedAxis(self):
        '''Get the widget for the linked axis.'''
        if not self.settings.linked:
            return None
        linked = self.lookupAxis(self.settings.linkedaxis)
        if linked is self:
            return None
        return linked

    def isLinear(self):
        return False

    def computePlottedRange(self, force=False):
        '''Use other axis to compute range.'''

        if self.docchangeset == self.document.changeset and not force:
            return

        therange = None

        linked = self.getLinkedAxis()
        fn = self.getFunction()
        if linked is not None and fn is not None:
            # compute our range from the linked axis
            linked.computePlottedRange()
            try:
                therange = fn(N.array(linked.plottedrange)) * N.ones(2)
            except Exception as e:
                self.logError(e)
            if not N.all( N.isfinite(therange) ):
                therange = None

        axis.Axis.computePlottedRange(
            self, force=force, overriderange=therange)

    def _orderCoordinates(self):
        '''Put coordinates in correct order for linear interpolation.'''

        if len(self.graphcoords) == 0:
            self.graphcoords = None
            return

        if self.graphcoords[0] > self.graphcoords[-1]:
            # order must be increasing (for forward conversion)
            self.graphcoords = self.graphcoords[::-1]
            self.pixcoords = self.pixcoords[::-1]
        if self.pixcoords_inv[0] > self.pixcoords_inv[-1]:
            # likewise increasing order for inverse
            self.pixcoords_inv = self.pixcoords_inv[::-1]
            self.graphcoords_inv = self.graphcoords_inv[::-1]

    def _updateLinkedAxis(self, bounds, fraccoords):
        '''Calculate coordinate conversion for linked axes.'''

        link = self.getLinkedAxis()
        if link is None:
            return

        # To do the inverse calculation, we define a grid of pixel
        # values.  We need some sensitivity outside the axis range to
        # get angles of lines correct. We start with fractional graph
        # coordinates to translate to the other axis coordinates.

        # coordinate values on the other axis
        try:
            linkwidth = link.coordParr2-link.coordParr1
            linkorigin = link.coordParr1
            linkbounds = link.currentbounds
        except AttributeError:
            # if hasn't been initialised
            return

        # lookup what pixels are on linked axis in values
        linkpixcoords = fraccoords*linkwidth + linkorigin
        linkgraphcoords = link.plotterToGraphCoords(linkbounds, linkpixcoords)

        # flip round if coordinates reversed
        if linkgraphcoords[0] > linkgraphcoords[-1]:
            linkgraphcoords = linkgraphcoords[::-1]
            linkpixcoords = linkpixcoords[::-1]
            fraccoords = fraccoords[::-1]

        # Chop to range. This is rather messy as there are several
        # sets of coordinates to extend and chop: graph coordinates,
        # pixel coordinates and fractional coordinates.
        mint, maxt = self.getMinMaxT()
        if mint is not None and mint > linkgraphcoords[0]:
            mintpix = link.graphToPlotterCoords(linkbounds, N.array([mint]))
            sel = linkgraphcoords > mint
            linkgraphcoords = N.hstack((mint, linkgraphcoords[sel]))
            linkpixcoords = N.hstack((mintpix, linkpixcoords[sel]))
            frac = (mintpix - linkorigin) / linkwidth
            fraccoords = N.hstack((frac, fraccoords[sel]))
        if maxt is not None and maxt < linkgraphcoords[-1]:
            maxtpix = link.graphToPlotterCoords(linkbounds, N.array([maxt]))
            sel = linkgraphcoords < maxt
            linkgraphcoords = N.hstack((linkgraphcoords[sel], maxt))
            linkpixcoords = N.hstack((linkpixcoords[sel], maxtpix))
            frac = (maxtpix - linkorigin) / linkwidth
            fraccoords = N.hstack((fraccoords[sel], frac))

        try:
            ourgraphcoords = self.getFunction()(linkgraphcoords)
        except:
            return

        deltas = ourgraphcoords[1:] - ourgraphcoords[:-1]
        pos = N.all(deltas >= 0.)
        neg = N.all(deltas <= 0.)
        if (not pos and not neg) or (pos and neg):
            self.logError(_('Not a monotonic function'))
            return

        # Select only finite vals. We store _inv coords separately
        # as linear interpolation requires increasing values.
        f = N.isfinite(linkgraphcoords + ourgraphcoords)
        self.graphcoords = self.graphcoords_inv = ourgraphcoords[f]

        # This is true if the axis is plotting on the same graph in
        # the same direction. If this is the case, use our coordinates
        # directly.
        if ( link.settings.direction == self.settings.direction
             and link.currentbounds == bounds ):
            self.pixcoords = self.pixcoords_inv = linkpixcoords[f]
        else:
            # convert fractions to our coordinates
            self.pixcoords = self.pixcoords_inv = (
                fraccoords[f]*(self.coordParr2-self.coordParr1)+self.coordParr1)

        # put output coordinates in correct order
        self._orderCoordinates()

    def _updateFreeAxis(self, bounds, fraccoords):
        '''Calculate coordinates for a free axis.'''

        self.computePlottedRange()
        trange = self.invertFunctionVals(N.array(self.plottedrange))
        if trange is None:
            return

        tvals = fraccoords*(trange[1]-trange[0]) + trange[0]

        if tvals[0] > tvals[-1]:
            # simplifies below if t is in order
            tvals = tvals[::-1]
            fraccoords = fraccoords[::-1]

        # limit t to the range if given
        mint, maxt = self.getMinMaxT()
        if mint is not None and mint > tvals[0]:
            sel = tvals > mint
            minfrac = (mint - trange[0]) / (trange[1]-trange[0])
            fraccoords = N.hstack( (minfrac, fraccoords[sel]) )
            tvals = N.hstack( (mint, tvals[sel]) )
        if maxt is not None and maxt < tvals[-1]:
            sel = tvals < maxt
            maxfrac = (maxt - trange[0]) / (trange[1]-trange[0])
            fraccoords = N.hstack( (fraccoords[sel], maxfrac) )
            tvals = N.hstack( (tvals[sel], maxt) )

        try:
            ourgraphcoords = self.getFunction()(tvals)
        except Exception:
            return

        deltas = ourgraphcoords[1:] - ourgraphcoords[:-1]
        pos = N.all(deltas >= 0.)
        neg = N.all(deltas <= 0.)
        if (not pos and not neg) or (pos and neg):
            self.logError(_('Not a monotonic function'))
            return

        # Select only finite vals. We store _inv coords separately
        # as linear interpolation requires increasing values.
        f = N.isfinite(ourgraphcoords)
        self.graphcoords = self.graphcoords_inv = ourgraphcoords[f]
        self.pixcoords = self.pixcoords_inv = (
            fraccoords[f]*(self.coordParr2-self.coordParr1) + self.coordParr1 )

        # put output coordinates in correct order
        self._orderCoordinates()

    def updateAxisLocation(self, bounds, otherposition=None,
                           lowerupperposition=None):
        '''Calculate conversion from pixels to axis values.'''

        axis.Axis.updateAxisLocation(
            self, bounds, otherposition=otherposition,
            lowerupperposition=lowerupperposition)

        if ( self.boundschangeset == self.document.changeset and
             bounds == self.cachedbounds ):
            # don't recalculate unless document updated or bounds changes
            return

        self.cachedbounds = list(bounds)
        self.boundschangeset = self.document.changeset

        self.graphcoords = None

        # fractional coordinate grid to evaluate functions
        fraccoords = N.hstack((
            N.linspace(-10., -2.0, 5, endpoint=False),
            N.linspace(-2.0, -0.2, 25, endpoint=False),
            N.linspace(-0.2, +1.2, 250, endpoint=False),
            N.linspace(+1.2, +3.0, 25, endpoint=False),
            N.linspace(+3.0, +11., 5, endpoint=False)
        ))

        if self.isLinked():
            self._updateLinkedAxis(bounds, fraccoords)
        else:
            self._updateFreeAxis(bounds, fraccoords)

    def _linearInterpolWarning(self, vals, xcoords, ycoords):
        '''Linear interpolation, giving out of bounds warning.'''
        if N.any(vals < xcoords[0]) or N.any(vals > xcoords[-1]):
            self.document.log(
                _('Warning: values exceed bounds in axis-function'))

        return N.interp(vals, xcoords, ycoords)

    def _graphToPlotter(self, vals):
        '''Override normal axis graph->plotter coords to do lookup.'''
        if self.graphcoords is None:
            return axis.Axis._graphToPlotter(self, vals)
        else:
            return self._linearInterpolWarning(
                vals, self.graphcoords, self.pixcoords)

    def plotterToGraphCoords(self, bounds, vals):
        '''Override normal axis plotter->graph coords to do lookup.'''
        if self.graphcoords is None:
            return axis.Axis.plotterToGraphCoords(self, bounds, vals)
        else:
            self.updateAxisLocation(bounds)

            return self._linearInterpolWarning(
                vals, self.pixcoords_inv, self.graphcoords_inv)

# allow the factory to instantiate the widget
document.thefactory.register( AxisFunction )



# consolewindow.py
# a python-like qt console

#    Copyright (C) 2003 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import codeop
import traceback
import sys

from .. import qtall as qt
from .. import document
from .. import utils
from .. import setting

# TODO - command line completion

def _(text, disambiguation=None, context='ConsoleWindow'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class _Writer:
    """ Class to behave like an output stream. Pipes input back to
    the specified function."""

    def __init__(self, function):
        """Set the function output is sent to."""
        self.function = function

    def write(self, text):
        """Send text to the output function."""
        self.function(text)

    def flush(self):
        """Does nothing as yet."""
        pass

class _Reader:
    """Fake reading input stream object."""
    def read(self, *args):
        raise IOError('Interactive input not supported')
    def readline(self, *args):
        raise IOError('Interactive input not supported')

class _CommandEdit(qt.QLineEdit):
    """ A special class to allow entering of the command line.

    emits sigEnter if the return key is pressed, and returns command
    The edit control has a history (press up and down keys to access)
    """

    sigEnter = qt.pyqtSignal(str)

    def __init__(self, *args):
        qt.QLineEdit.__init__(self, *args)
        self.history = []
        self.history_posn = 0
        self.entered_text = ''

        self.returnPressed.connect(self.slotReturnPressed)

        self.setToolTip(_("Input a python expression here and press enter"))

    def slotReturnPressed(self):
        """ Called if the return key is pressed in the edit control."""

        # retrieve the text
        command = self.text()
        self.setText("")

        # keep the command for history
        self.history.append(command)
        self.history_posn = len(self.history)
        self.entered_text = ''

        # tell the console we have a command
        self.sigEnter.emit(command)

    historykeys = (qt.Qt.Key_Up, qt.Qt.Key_Down)

    def keyPressEvent(self, key):
        """ Overridden to handle history. """

        qt.QLineEdit.keyPressEvent(self, key)
        code = key.key()

        # check whether one of the "history keys" has been pressed
        if code in _CommandEdit.historykeys:

            # look for the next or previous history item which our current text
            # is a prefix of
            if self.isModified():
                text = self.text()
                self.history_posn = len(self.history)
            else:
                text = self.entered_text

            if code == qt.Qt.Key_Up:
                step = -1
            elif code == qt.Qt.Key_Down:
                step = 1

            newpos = self.history_posn + step

            while True:
                if newpos >= len(self.history):
                    break
                if newpos < 0:
                    return
                if self.history[newpos].startswith(text):
                    break

                newpos += step

            if newpos >= len(self.history):
                # go back to whatever the user had typed in
                self.history_posn = len(self.history)
                self.setText(self.entered_text)
                return

            # found a relevant history item
            self.history_posn = newpos

            # user has modified text since last set
            if self.isModified():
                self.entered_text = text

            # replace the text in the control
            text = self.history[ self.history_posn ]
            self.setText(text)

introtext=_('''Welcome to <b><font color="purple">Veusz %s</font></b> --- a scientific plotting application.<br>
Copyright \u00a9 2003-2022 Jeremy Sanders &lt;jeremy@jeremysanders.net&gt; and contributors.<br>
Veusz comes with ABSOLUTELY NO WARRANTY. Veusz is Free Software, and you are<br>
welcome to redistribute it under certain conditions. Enter "GPL()" for details.<br>
This window is a Python command line console and acts as a calculator.<br>
''') % utils.version()

class ConsoleWindow(qt.QDockWidget):
    """ A python-like qt console."""

    def __init__(self, thedocument, *args):
        qt.QDockWidget.__init__(self, *args)
        self.setWindowTitle(_("Console - Veusz"))
        self.setObjectName("veuszconsolewindow")

        # arrange sub-widgets in a vbox
        self.vbox = qt.QWidget()
        self.setWidget(self.vbox)
        vlayout = qt.QVBoxLayout(self.vbox)
        s = vlayout.contentsMargins().left()//4
        vlayout.setContentsMargins(s,s,s,s)
        vlayout.setSpacing(s)

        # start an interpreter instance to the document
        self.interpreter = document.CommandInterpreter(thedocument)
        self.document = thedocument

        # streams the output/input goes to/from
        self.con_stdout = _Writer(self.output_stdout)
        self.con_stderr = _Writer(self.output_stderr)
        self.con_stdin = _Reader()

        self.interpreter.setFiles(
            self.con_stdout, self.con_stderr, self.con_stdin)
        self.stdoutbuffer = ""
        self.stderrbuffer = ""

        # (mostly) hidden notification
        self._hiddennotify = qt.QLabel()
        vlayout.addWidget(self._hiddennotify)
        self._hiddennotify.hide()

        # the output from the console goes here
        self._outputdisplay = qt.QTextEdit()
        self._outputdisplay.setReadOnly(True)
        self._outputdisplay.insertHtml( introtext )
        vlayout.addWidget(self._outputdisplay)

        self._hbox = qt.QWidget()
        hlayout = qt.QHBoxLayout(self._hbox)
        hlayout.setContentsMargins(0,0,0,0)
        vlayout.addWidget(self._hbox)

        self._prompt = qt.QLabel(">>>")
        hlayout.addWidget(self._prompt)

        # where commands are typed in
        self._inputedit = _CommandEdit()
        hlayout.addWidget(self._inputedit)
        self._inputedit.setFocus()

        # keep track of multiple line commands
        self.command_build = ''

        # get called if enter is pressed in the input control
        self._inputedit.sigEnter.connect(self.slotEnter)
        # called if document logs something
        thedocument.sigLog.connect(self.slotDocumentLog)

    def _makeTextFormat(self, cursor, color):
        fmt = cursor.charFormat()

        if color is not None:
            brush = qt.QBrush(color)
            fmt.setForeground(brush)
        else:
            # use the default foreground color
            fmt.clearForeground()

        return fmt

    def appendOutput(self, text, style):
        """Add text to the tail of the error log, with a specified style"""
        if style == 'error':
            color = setting.settingdb.color('error')
        elif style == 'command':
            color = setting.settingdb.color('command')
        else:
            color = None

        cursor = self._outputdisplay.textCursor()
        cursor.movePosition(qt.QTextCursor.End)
        cursor.insertText(text, self._makeTextFormat(cursor, color))
        self._outputdisplay.setTextCursor(cursor)
        self._outputdisplay.ensureCursorVisible()

    def runFunction(self, func):
        """Execute the function within the console window, trapping
        exceptions."""

        # preserve output streams
        saved = sys.stdout, sys.stderr, sys.stdin
        sys.stdout, sys.stderr, sys.stdin = (
            self.con_stdout, self.con_stderr, self.con_stdin)

        # catch any exceptions, printing problems to stderr
        with self.document.suspend():
            try:
                func()
            except:
                # print out the backtrace to stderr
                info = sys.exc_info()
                backtrace = traceback.format_exception(*info)
                for line in backtrace:
                    sys.stderr.write(line)

        # return output streams
        sys.stdout, sys.stderr, sys.stdin = saved

    def checkVisible(self):
        """If this window is hidden, show it, then hide it again in a few
        seconds."""
        if self.isHidden():
            self._hiddennotify.setText(_(
                "This window will shortly disappear. "
                "You can bring it back by selecting "
                "View, Windows, Console Window on the menu."))
            qt.QTimer.singleShot(5000, self.hideConsole)
            self.show()
            self._hiddennotify.show()

    def hideConsole(self):
        """Hide window and notification widget."""
        self._hiddennotify.hide()
        self.hide()

    def output_stdout(self, text):
        """ Write text in stdout font to the log."""
        self.checkVisible()
        self.appendOutput(text, 'normal')

    def output_stderr(self, text):
        """ Write text in stderr font to the log."""
        self.checkVisible()
        self.appendOutput(text, 'error')

    def insertTextInOutput(self, text):
        """ Inserts the text into the log."""
        self.appendOutput(text, 'normal')

    def slotEnter(self, command):
        """ Called if the return key is pressed in the edit control."""

        newc = self.command_build + '\n' + command

        # check whether command can be compiled
        # c set to None if incomplete
        try:
            comp = codeop.compile_command(newc)
        except Exception:
            # we want errors to be caught by self.interpreter.run below
            comp = 1

        # which prompt?
        prompt = '>>>'
        if self.command_build != '':
            prompt = '...'

        # output the command in the log pane
        self.appendOutput('%s %s\n' % (prompt, command), 'command')

        # are we ready to run this?
        if comp is None or (
                len(command) != 0 and
                len(self.command_build) != 0 and
                (command[0] == ' ' or command[0] == '\t')):
            # build up the expression
            self.command_build = newc
            # modify the prompt
            self._prompt.setText('...')
        else:
            # actually execute the command
            self.interpreter.run(newc)
            self.command_build = ''
            # modify the prompt
            self._prompt.setText('>>>')

    def slotDocumentLog(self, text):
        """Output information if the document logs something."""
        self.output_stderr(text + '\n')



#    Copyright (C) 2011 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Parameters for import routines."""

import sys
import copy

from .. import utils

class ImportingError(RuntimeError):
    """Common error when import fails."""

class ImportParamsBase:
    """Import parameters for the various imports.

    Parameters:
     filename: filename to import from
     linked: whether to link to file
     encoding: encoding for file
     prefix: prefix for output dataset names
     suffix: suffix for output dataset names
     tags: list of tags to apply to output datasets
     renames: dict map of names to renamed datasets
    """

    defaults = {
        'filename': None,
        'linked': False,
        'encoding': 'utf_8',
        'prefix': '',
        'suffix': '',
        'tags': None,
        'renames': None,
    }

    def __init__(self, **argsv):
        """Initialise the reader to import data from filename.
        """

        #  set defaults
        for k, v in self.defaults.items():
            setattr(self, k, v)

        # set parameters
        for k, v in argsv.items():
            if k not in self.defaults:
                raise ValueError("Invalid parameter %s" % k)
            setattr(self, k, v)

        # extra parameters to copy besides defaults
        self._extras = []

    def copy(self):
        """Make a copy of the parameters object."""

        newp = {}
        for k in list(self.defaults.keys()) + self._extras:
            newp[k] = getattr(self, k)
        return self.__class__(**newp)

class LinkedFileBase:
    """A base class for linked files containing common routines."""

    def __init__(self, params):
        """Save parameters."""
        self.params = params

    def createOperation(self):
        """Return operation to recreate self."""
        return None

    @property
    def filename(self):
        """Get filename."""
        return self.params.filename

    def _saveHelper(self, fileobj, cmd, fixedparams,
                    renameparams={}, relpath=None, extraargs={}):
        """Helper to write command to reload data.

        fileobj: file object to write to
        cmd: name of command to write
        fixedparams: list of parameters to list at start of command lines
        renameparams: optional map of params to command line params
        relpath: relative path for writing filename
        extraargs: other options to add to command line
        """

        p = self.params
        args = []

        # arguments without names at command start
        for par in fixedparams:
            if par == 'filename':
                v = self._getSaveFilename(relpath)
            else:
                v = getattr(p, par)
            args.append(utils.rrepr(v))

        # parameters key, values to put in command line
        plist = sorted(
            [(par, getattr(p, par)) for par in p.defaults] +
            list(extraargs.items())
        )

        for par, val in plist:
            if ( val and
                 (par not in p.defaults or p.defaults[par] != val) and
                 par not in fixedparams and
                 par != 'tags' ):

                if par in renameparams:
                    par = renameparams[par]
                args.append('%s=%s' % (par, utils.rrepr(val)))

        # write command using comma-separated list
        fileobj.write('%s(%s)\n' % (cmd, ', '.join(args)))

    def saveToFile(self, fileobj, relpath=None):
        """Save the link to the document file."""
        pass

    def _getSaveFilename(self, relpath):
        """Get filename to write to save file.
        If relpath is a string, write relative to path given
        """
        if relpath:
            f = utils.relpath(self.params.filename, relpath)
        else:
            f = self.filename
        # Here we convert backslashes in Windows to forward slashes
        # This is compatible, but also works on Unix/Mac
        if sys.platform == 'win32':
            f = f.replace('\\', '/')
        return f

    def _deleteLinkedDatasets(self, document):
        """Delete linked datasets from document linking to self.
        Returns tags for deleted datasets.
        """

        tags = {}
        for name, ds in list(document.data.items()):
            if ds.linked == self:
                tags[name] = document.data[name].tags
                document.deleteData(name)
        return tags

    def _moveReadDatasets(self, tempdoc, document, tags):
        """Move datasets from tempdoc to document if they do not exist
        in the destination.

        tags is a dict of tags for each dataset
        """

        read = []
        for name, ds in list(tempdoc.data.items()):
            if name not in document.data:
                ds.linked = self
                if name in tags:
                    ds.tags = tags[name]
                document.setData(name, ds)
                read.append(name)
        return read

    def reloadLinks(self, document):
        """Reload links using an operation"""

        # get the operation for reloading
        op = self.createOperation()(self.params)

        # load data into a temporary document
        tempdoc = document.__class__()

        try:
            tempdoc.applyOperation(op)
        except Exception as ex:
            # if something breaks, record an error and return nothing
            document.log(str(ex))

            # find datasets which are linked using this link object
            # return errors for them
            errors = dict(
                [(name, 1) for name, ds in document.data.items()
                 if ds.linked is self])
            return ([], errors)

        # delete datasets which are linked and imported here
        tags = self._deleteLinkedDatasets(document)
        # move datasets into document
        read = self._moveReadDatasets(tempdoc, document, tags)

        # return errors (if any)
        errors = op.outinvalids

        return (read, errors)

class OperationDataImportBase:
    """Default useful import class."""

    def __init__(self, params):
        self.params = params

    def doImport(self, document):
        """Do import, override this.
        Set outdatasets
        """

    def addCustoms(self, document, customs):
        """Optionally, add the customs return by plugins to document."""

        type_attrs = {
            'import': 'def_imports',
            'color': 'def_colors',
            'colormap': 'def_colormaps',
            'constant': 'def_definitions',
            'function': 'def_definitions',
            'definition': 'def_definitions',
        }

        if len(customs) > 0:
            doceval = document.evaluate
            self.oldcustoms = [
                copy.deepcopy(doceval.def_imports),
                copy.deepcopy(doceval.def_definitions),
                copy.deepcopy(doceval.def_colors),
                copy.deepcopy(doceval.def_colormaps)]

            # FIXME: inefficient for large number of definitions
            for item in customs:
                ctype, name, val = item
                clist = getattr(doceval, type_attrs[ctype])
                for idx, (cname, cval) in enumerate(clist):
                    if cname == name:
                        clist[idx][1] = val
                        break
                else:
                    clist.append([name, val])

            doceval.update()

    def do(self, document):
        """Do import."""

        # list of returned dataset names
        self.outnames = []
        # map of names to datasets
        self.outdatasets = {}
        # list of returned custom variables
        self.outcustoms = []
        # invalid conversions
        self.outinvalids = {}

        # remember datasets in document for undo
        self.oldcustoms = None

        # do actual import
        retn = self.doImport()

        # these are custom values returned from the plugin
        if self.outcustoms:
            self.addCustoms(document, self.outcustoms)

        # handle tagging/renaming
        for name, ds in list(self.outdatasets.items()):
            if self.params.tags:
                ds.tags.update(self.params.tags)
            if self.params.renames and name in self.params.renames:
                del self.outdatasets[name]
                self.outdatasets[self.params.renames[name]] = ds

        # only remember the parts we need
        self.olddatasets = [
            (n, document.data.get(n)) for n in self.outdatasets ]

        self.olddatasets = []
        for name, ds in self.outdatasets.items():
            self.olddatasets.append( (name, document.data.get(name)) )
            document.setData(name, ds)

        self.outnames = sorted(self.outdatasets)

        return retn

    def undo(self, document):
        """Undo import."""

        # put back old datasets
        for name, ds in self.olddatasets:
            if ds is None:
                document.deleteData(name)
            else:
                document.setData(name, ds)

        # for custom definitions
        if self.oldcustoms is not None:
            doceval = document.evaluate
            doceval.def_imports = self.oldcustoms[0]
            doceval.def_definitions = self.oldcustoms[1]
            doceval.def_colors = self.oldcustoms[2]
            doceval.def_colormaps = self.oldcustoms[3]
            doceval.update()



#    Copyright (C) 2008 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import sys
import struct
import socket
import pickle

from . import qtall as qt
from .windows.simplewindow import SimpleWindow
from . import document
from . import setting

"""Program to be run by embedding interface to run Veusz commands."""

# embed.py module checks this is the same as its version number
API_VERSION = 2

class EmbeddedClient:
    """An object for each instance of embedded window with document."""

    def __init__(self, title, doc=None, hidden=False):
        """Construct window with title given."""

        self.window = SimpleWindow(title, doc=doc)
        if not hidden:
            self.window.show()
        self.document = self.window.document
        self.plot = self.window.plot
        # use time based checking by default
        self.plot.setTimeout(250)
        self.ci = document.CommandInterpreter(self.document)
        self.ci.addCommand('Close', self.cmdClose)
        self.ci.addCommand('Zoom', self.cmdZoom)
        self.ci.addCommand('EnableToolbar', self.cmdEnableToolbar)
        self.ci.addCommand('ForceUpdate', self.cmdForceUpdate)
        self.ci.addCommand('GetClick', self.cmdGetClick)
        self.ci.addCommand('ResizeWindow', self.cmdResizeWindow)
        self.ci.addCommand('SetUpdateInterval', self.cmdSetUpdateInterval)
        self.ci.addCommand('MoveToPage', self.cmdMoveToPage)
        self.ci.addCommand('IsClosed', self.cmdIsClosed)
        self.ci.addCommand('SetAntiAliasing', self.cmdSetAntiAliasing)
        self.ci.addCommand('Wipe', self.cmdWipe)
        self.ci.addCommand('_apiVersion', self.cmd_apiVersion)

        setting.transient_settings['unsafe_mode'] = True

        self.document.sigLog.connect(self.logEmitted)

    def logEmitted(self, msg):
        """Write anything logged to stderr."""
        sys.stderr.write(msg + '\n')

    def cmdClose(self):
        """Close()

        Close this window."""
        self.window.close()

        self.document = None
        self.window = None
        self.plot = None
        self.ci = None

    def cmdIsClosed(self):
        """IsClosed()

        Return whether window is still open."""
        return not self.window.isVisible()

    def cmd_apiVersion(self):
        """Get internal API version."""
        return API_VERSION

    def cmdZoom(self, zoom):
        """Zoom(zoom)

        Set the plot zoom level:
        This is a number to for the zoom from 1:1 or
        'page': zoom to page
        'width': zoom to fit width
        'height': zoom to fit height
        """
        self.window.setZoom(zoom)

    def cmdSetAntiAliasing(self, ison):
        """SetAntiAliasing(zoom)

        Enables or disables anti aliasing.
        """
        self.window.setAntiAliasing(ison)

    def cmdEnableToolbar(self, enable=True):
        """EnableToolbar(enable=True)

        Enable the toolbar in this plotwindow.
        if enable is False, disable it.
        """
        self.window.enableToolbar(enable)

    def cmdForceUpdate(self):
        """ForceUpdate()

        Forces an update of the plot window.
        """
        self.plot.actionForceUpdate()

    def cmdGetClick(self):
        """GetClick()

        Return a clicked point. The user can click a point on the graph

        This returns a list of tuples containing items for each axis in
        the clicked region:
         (axisname, valonaxis)
        where axisname is the full name of an axis
        valonaxis is value clicked along the axis

        [] is returned if no axes span the clicked region
        """
        return self.plot.getClick()

    def cmdResizeWindow(self, width, height):
        """ResizeWindow(width, height)

        Resize the window to be width x height pixels."""
        self.window.resize(width, height)

    def cmdSetUpdateInterval(self, interval):
        """SetUpdateInterval(interval)

        Set graph update interval.
        interval is in milliseconds (ms)
        set to zero to disable updates
        set to -1 to update when document changes
        default interval is 250ms
        """
        self.plot.setTimeout(interval)

    def cmdMoveToPage(self, pagenum):
        """MoveToPage(pagenum)

        Tell window to show specified pagenumber (starting from 1).
        """
        self.plot.setPageNumber(pagenum-1)

    def cmdWipe(self):
        """Wipe the current document."""
        self.document.wipe()

class EmbedApplication(qt.QApplication):
    """Application to run remote end of embed connection.

    Commands are sent over stdin, with responses sent to stdout
    """

    # lengths of lengths sent to application
    cmdlenlen = struct.calcsize('<I')

    def __init__(self, thesocket, args):
        qt.QApplication.__init__(self, args)
        self.socket = thesocket

        # listen to commands on the socket
        self.notifier = qt.QSocketNotifier(
            self.socket.fileno(), qt.QSocketNotifier.Read)
        self.notifier.activated.connect(self.slotDataToRead)
        self.notifier.setEnabled(True)

        # keep track of clients (separate veusz documents)
        self.clients = {}
        self.clientcounter = 0

    @staticmethod
    def readLenFromSocket(thesocket, length):
        """Read length bytes from socket."""
        s = b''
        while len(s) < length:
            s += thesocket.recv(length-len(s))
        return s

    @staticmethod
    def writeToSocket(thesocket, data):
        """Write to socket until all data written."""
        count = 0
        while count < len(data):
            count += thesocket.send(data[count:])

    @staticmethod
    def readCommand(thesocket):
        # get length of packet
        length = struct.unpack('<I', EmbedApplication.readLenFromSocket(
                thesocket, EmbedApplication.cmdlenlen))[0]
        # unpickle command and arguments
        temp = EmbedApplication.readLenFromSocket(thesocket, length)
        return pickle.loads(temp)

    def makeNewClient(self, title, doc=None, hidden=False):
        """Make a new client window."""
        client = EmbeddedClient(title, doc=doc, hidden=hidden)
        self.clients[self.clientcounter] = client
        # return new number and list of commands and docstrings
        retfuncs = []
        for name, cmd in client.ci.cmds.items():
            retfuncs.append( (name, cmd.__doc__) )

        retval = self.clientcounter, retfuncs
        self.clientcounter += 1
        return retval

    def writeOutput(self, output):
        """Send output back to embed process."""
        # format return data
        outstr = pickle.dumps(output)

        # send return data to stdout
        self.writeToSocket( self.socket, struct.pack('<I', len(outstr)) )
        self.writeToSocket( self.socket, outstr )

    def finishRemote(self):
        """Clean up on exit."""
        self.notifier.setEnabled(False)
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        self.closeAllWindows()
        self.quit()

    def slotDataToRead(self, socketfd):
        """Call routine to read data from remote socket."""
        try:
            self.readFromSocket()
        except socket.error:
            # exit if problem
            self.finishRemote()

    def readFromSocket(self):
        self.notifier.setEnabled(False)
        self.socket.setblocking(1)

        # unpickle command and arguments
        window, cmd, args, argsv = self.readCommand(self.socket)

        if cmd == '_NewWindow':
            retval = self.makeNewClient(args[0], hidden=argsv['hidden'])
        elif cmd == '_Quit':
            # exit client
            retval = None
        elif cmd == '_NewWindowCopy':
            # sets the document of this window to be the same as the
            # one specified
            retval = self.makeNewClient(
                args[0],
                doc=self.clients[args[1]].document,
                hidden=argsv['hidden']
            )
        else:
            interpreter = self.clients[window].ci

            # window commands
            try:
                if cmd not in interpreter.cmds:
                    raise AttributeError("No Veusz command %s" % cmd)

                retval = interpreter.cmds[cmd](*args, **argsv)
            except Exception as e:
                retval = e

        self.writeOutput(retval)

        # do quit after if requested
        if cmd == '_Quit':
            self.finishRemote()
            return

        self.socket.setblocking(0)
        self.notifier.setEnabled(True)

def runremote():
    """Run remote end of embedding module."""
    # get connection parameters
    params = sys.stdin.readline().split()

    if params[0] == 'unix':
        # talk to existing unix domain socket
        listensocket = socket.fromfd(
            int(params[1]),
            socket.AF_UNIX,
            socket.SOCK_STREAM
        )

    elif params[0] == 'internet':
        # talk to internet port
        listensocket = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        listensocket.connect( (params[1], int(params[2])) )

    # get secret from stdin and send back to socket
    # this is a security check
    secret = sys.stdin.readline().encode('ascii')
    EmbedApplication.writeToSocket(listensocket, secret)

    # finally start listening application
    app = EmbedApplication(listensocket, [])
    app.setQuitOnLastWindowClosed(False)
    app.exec_()
